/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "ainet_pass.hpp"
#include "immune_system.hpp"
#include "immune_system.tpp"

//pass_repertoire::registerer ainet_pass::m_reg(new ainet_pass());

param_info 
ainet_pass::m_param_info[ainet_pass::m_param_info_size]=
{
	param_info(1, 10, "Number of Iterations", "Number of passes to apply"),
	param_info(10, 255, "Sigma_S", "Threshold for network suppression"),
	param_info(10, 255, "Sigma_D", "Threshold for clonal suppression"),
	param_info(3, 10, "N", "Number of best responsive antibodies to select"),
	param_info(0.3, 1.0, "Zeta (%)", "Percentage of reselected best antibodies")
};

/// Input iterator that will return copies of a given object
/// a given number of times before becoming an end iterator
template<typename T>
struct clone_iterator
{
	/// Takes prototype and intial prototype count
	clone_iterator(T const& proto, std::size_t num): prototype(proto),
		count(num)
	{}

	T prototype;
	std::size_t count;

	T& operator*(){return prototype;}

	clone_iterator& operator++(){count--; return *this;}

	bool operator==(clone_iterator const& lhs)
	{ return lhs.prototype==prototype && count==lhs.count; }

	bool operator!=(clone_iterator const& lhs)
	{ return ! this->operator==(lhs); }

};

/// Enable the use of immune_system with histogram::gray_level_type as
/// antibodies
template<>
struct antibody_traits<histogram::gray_level_type>
{
	typedef histogram::gray_level_type antibody_type;
	typedef histogram::gray_level_type distance_type;

	static void mutate(antibody_type* cl, antibody_type ag,
			double percent)
	{
		*cl=*cl-antibody_type(percent*(*cl-ag));
	}

	static distance_type distance(antibody_type t, antibody_type u)
	{ return std::abs(t-u); }

	static void print(antibody_type,std::FILE*){}
};

void ainet_pass::print() const
{
	std::printf("%s %d,%d,%d,%d,%f \n",
			label(),
			parameter_value(0).int_value(),
			parameter_value(1).int_value(),
			parameter_value(2).int_value(),
			parameter_value(3).int_value(),
			parameter_value(4).double_value());
}

void ainet_pass::operator() (histogram const& hist,
			centroid_set& centroids,
			progress_indicator* ind) const
{
	int const nr_passes=this->pass::parameter_value(0).int_value();
	int const sigma_s=this->pass::parameter_value(1).int_value();
	int const sigma_d=this->pass::parameter_value(2).int_value();
	int const n=this->pass::parameter_value(3).int_value();
	double const zeta_percent=this->pass::parameter_value(4).double_value();

	immune_system<histogram::gray_level_type> is(sigma_s, sigma_d);
	is.n=n;
	is.zeta_percent=zeta_percent;

	double total_pixels=0;
	for(histogram::const_iterator it=hist.begin();
			it!=hist.end(); ++it)
		total_pixels+=*it;

	// Intialize the immune system (artificially) with given centroids
	for(centroid_set::iterator it=centroids.begin();
			it!=centroids.end(); ++it){
		is.insert(histogram::gray_level_type(*it));
	}
	
	int pass_count=nr_passes;
	double pixels_so_far=0;
	while(pass_count--){
		pixels_so_far=0;

		// Expose the immune system to all gray levels
		for(histogram::const_iterator it=hist.begin();
				it!=hist.end(); ++it){
			histogram::gray_level_type const level=
				it-hist.begin();

			clone_iterator<histogram::gray_level_type> beg(level, *it);
			clone_iterator<histogram::gray_level_type> end(level, 0);
			is.stimulate(beg, end);

			pixels_so_far+=*it;
			ind->operator() (pixels_so_far/total_pixels);
		}
	}// while pass_count--
	
	// Replace old centroids with those of the immune_system<>
	// (in one shot! God bless the STL!)
	centroids.assign(is.begin(), is.end());	
}

