/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "dyn_pso_utils.hpp"
#include "dyn_pso.hpp"


dyn_particle::dyn_particle (int dyn_particle_size, double seed, int velocity_range):Pini(seed)
{

    for (int i=0; i< dyn_particle_size; ++i)
    {

        double r=myrandom(0)();//generate U(0,1)
        if (r>= seed)
            current_position.push_back(0);
        else
            current_position.push_back(1);
    }
    generate_n(back_inserter(velocity),dyn_particle_size,myrandom(velocity_range));
    copy(current_position.begin(), current_position.end(), back_inserter(best_position));
}


void dyn_particle::print()
{
    std::cout<<"Current position: "<<std::endl;
    for(unsigned i=0 ;i<current_position.size(); ++i)
    {
        std::cout<<current_position[i]<<" ";
    }
    std::cout<<std::endl;
    std::cout<<"current position cost: "<<cost<<std::endl;
    std::cout<<"Best position: "<<std::endl;
    for(unsigned i=0 ;i<best_position.size(); ++i)
    {
        std::cout<<best_position[i]<<" ";
    }
    std::cout<<std::endl;
    std::cout<<"Best position cost: "<<best_cost<<std::endl;
    std::cout<<"Velocity: "<<std::endl;
    for(unsigned i=0 ;i<velocity.size(); ++i)
    {
        std::cout<<velocity[i]<<" ";
    }
    std::cout<<std::endl;
    std::cout<<"------------------------------------------------------"<<std::endl;

}


dyn_pso_swarm::dyn_pso_swarm(double _w, double _c1, double _c2, int _s, int _ps,
             double _Pini, int _vr, std::vector<double> & _centroids, std::vector<cluster>& _hist,
             double (*_fini)(std::vector<bool>&, std::vector<double>&, std::vector<cluster>&))
        :w(_w),c1(_c1), c2(_c2),s(_s), ps(_ps),Pini(_Pini), vr(_vr), centroids(_centroids),
        hist(_hist), fini(_fini)

{

    dyn_particle p(ps,Pini,vr); //Init dyn_particle
    for (int i=0;i<s;++i)
    {
        dyn_particle p(ps,Pini,vr);
        p.set_cost(fini(p.get_current_position(),centroids,hist));
        p.set_best_cost(p.get_cost());
        p_repertoire.push_back(p);

    }

    generate_n(back_inserter(r1),psize(),myrandom(0));
    generate_n(back_inserter(r2),psize(),myrandom(0));

    //set the best dyn_particle of the dyn_pso_swarm to the first one
    best_dyn_particle_cost=p_repertoire[0].get_cost();

    best_dyn_particle_position.clear();
    copy(p_repertoire[0].get_best_position().begin(),
         p_repertoire[0].get_best_position().end(),
         back_inserter(best_dyn_particle_position));

    for(std::vector<dyn_particle>::iterator it=p_repertoire.begin();it!=p_repertoire.end();++it)
    {
        if(it->get_best_cost()<best_dyn_particle_cost)
        {
            //Update the global best position
            best_dyn_particle_position.clear();
            copy(it->get_best_position().begin(),it->get_best_position().end(),
                 back_inserter(best_dyn_particle_position));

            //Update the global best cost
            best_dyn_particle_cost=it->get_best_cost();
        }
    }

}


void dyn_pso_swarm::update_dyn_particle(dyn_particle& p)
{
    //Updating dyn_particle velocity
    for(int i=0;i<psize();++i)
    {
        //Velocity update function
        p.get_velocity()[i]=w*p.get_velocity()[i]+
                            c1*r1[i]*(p.get_best_position()[i]-p.get_current_position()[i])+
                            c2*r2[i]*(get_best_dyn_particle_position()[i]-p.get_current_position()[i]);

    }

    //updating the current position
    //transform(p.current().begin(), p.current().end(), p.speed().begin(),p.current().begin(), std::plus<int>());
    for(int i=0;i<psize();++i)
    {
        double r=myrandom(0)();
        double sigmoid=1/(1+exp(-p.get_velocity()[i]));
        if (r>= sigmoid)
            p.get_current_position()[i]=0;
        else
            p.get_current_position()[i]=1;
    }
    p.set_cost(fini(p.get_current_position(), centroids, hist));

    //Updating the best position of a dyn_particle so far
    if(p.get_cost()<p.get_best_cost())
    {
        p.get_best_position().clear();
        copy(p.get_current_position().begin(),p.get_current_position().end(),
             back_inserter(p.get_best_position()));
        p.set_best_cost(p.get_cost());
    }


}

void dyn_pso_swarm::update_swarm()
{
    //Updating all the dyn_particles of the dyn_pso_swarm
    for(int i=0; i<s; ++i)
    {
        update_dyn_particle(p_repertoire[i]);
    }

    //Updating the best dyn_particle postition( the optimal solution so far)
    for(std::vector<dyn_particle>::iterator it=p_repertoire.begin();it!=p_repertoire.end();++it)
    {
        if(it->get_best_cost()<best_dyn_particle_cost)
        {
            //Update the global best position
            best_dyn_particle_position.clear();
            copy(it->get_best_position().begin(),it->get_best_position().end(),
                 back_inserter(best_dyn_particle_position));

            //Update the global best cost
            best_dyn_particle_cost=it->get_best_cost();
        }
    }
}
;
