/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef DYN_PSO_HPP
#define DYN_PSO_HPP

#include<iostream>
#include<vector>
#include<algorithm>
#include "dyn_pso_utils.hpp"

//Represents an individual (dyn_particle) of a dyn_pso_swarm
class dyn_particle
{
private:
    std::vector<bool> current_position;//current position of a dyn_particle
    std::vector<bool> best_position; //the best position reacher by a dyn_particle so far
    std::vector<double> velocity; //the speed vector of a dyn_particle
    double Pini; //Random value U(0,1) used to initialize dyn_particle
    double cost; //The fitness value of a dyn_particle
    double best_cost;
public:

    //First constructor which make a dyn_particle of dyn_particle_size size
    //seed is used to set a dyn_particle components to 1 or 0
    //velocity_range is used to initialize each component of velocity vector
    //in the rage [- velocity_range, +velocity_range]
    dyn_particle (int dyn_particle_size, double seed, int velocity_range);

    //print a dyn_particle
    void print();

    //yields the current position of the dyn_particle
    std::vector<bool>& get_current_position()
    {
        return current_position;
    }

    //yields the velocity of the dyn_particle
    std::vector<double>& get_velocity()
    {
        return velocity;
    }

    //yields the best position of the dyn_particle so far
    std::vector<bool>& get_best_position()
    {
        return best_position;
    }

    //sets a fitness value to a dyn_particle
    void set_cost(double c)
    {
        cost=c;
    }

    //yields a fitness value from a dyn_particle
    double get_cost()
    {
        return cost;
    }

    //sets the best fitness value of a dyn_particle
    void set_best_cost(double c)
    {
        best_cost=c;
    }

    //gets the best fitness value from a dyn_particle
    double get_best_cost()
    {
        return best_cost;
    }
}
;// class dyn_particle

//This class holds all the dyn_particles of the dyn_pso_swarm and the globale characteritics
//needed for the evolutionary process
class dyn_pso_swarm
{

private:
    std::vector<dyn_particle>  p_repertoire;
    std::vector<bool> best_dyn_particle_position;
    std::vector<double> r1; //the r1->U(0,1) parameters to update dyn_particle velocity
    std::vector<double> r2; // the r2->U(0,1) parameter to update dyn_particle velocity
    double w;  //w is the inertia
    double c1,c2; // acceleration constants
    int s; //dyn_pso_swarm size
    int ps; //dyn_particle size

    double Pini;//value U(0,1) to initialize dyn_particles
    int vr; //Velocity range [-vr,+vr]
    double best_dyn_particle_cost;
    //Image specifications
    std::vector<double>& centroids;
    std::vector<cluster>& hist;
    double (*fini)(std::vector<bool>&, std::vector<double>&,std::vector<cluster>& );

public:

    //The arguments of the constructor are in this order: inertia, first constant,
    //second constant, dyn_pso_swarm size, dyn_particle size, seed to initialize dyn_particles,
    //velocity range, reference to centroids vector, reference to the clustering
    //environement(histogram), the last one is a pointer to fuction to calculate
    //the fitness of a dyn_particle X[i] in the dyn_pso_swarm; its first argument is a reference
    //to a dyn_particle which we want to calculate its fitness, the second is a centroids vector,
    //and the last argument is a reference to the histogram.

    dyn_pso_swarm(double inertia, double first_constant, double second_constant,
          int dyn_pso_swarm_size, int dyn_particle_size,double _Pini, int velocity_range,
          std::vector<double> & centroids,std::vector<cluster>& histogram,
          double (*_fini)(std::vector<bool>&, std::vector<double>&,std::vector<cluster>&));

    //This member function is to update a dyn_particle of the dyn_pso_swarm position (toward the best one)
    //according to the previous ones and to the neighbourhood
    void update_dyn_particle(dyn_particle& p);

    //This member function is to update the entire dyn_pso_swarm
    void update_swarm();


    //Return the dyn_pso_swarm size
    const int ssize()
    {
        return s;
    }

    //Return the dyn_particle size
    const int psize()
    {
        return ps;
    }

    //Returns a ith dyn_particles
    dyn_particle& operator[] (int i)
    {
        return p_repertoire[i];
    }

    double get_best_dyn_particle_cost()
    {
        return best_dyn_particle_cost;
    }

    //sets the cost of the best dyn_particle of the dyn_pso_swarm
    void set_best_dyn_particle_cost(double b)
    {
        best_dyn_particle_cost=b;
    }

    //gets the best dyn_particle of the dyn_pso_swarm
    std::vector<bool>& get_best_dyn_particle_position()
    {
        return best_dyn_particle_position;
    }

    //sets the best dyn_particle of the dyn_pso_swarm
    void set_best_dyn_particle_position(std::vector<bool>& p)
    {
        best_dyn_particle_position.clear();
        copy(p.begin(), p.end(), back_inserter(best_dyn_particle_position));
    }
}//class dyn_pso_swarm
;

#endif // DYN_PSO_HPP
