/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <cmath>
#include <cstdio>

#include "fcm_pass.hpp"
#include "histogram.hpp"
#include "classification.hpp"

/// Definition of parameter info array (contains an element per parameter)
param_info 
fcm_pass::m_param_info[fcm_pass::m_param_info_size]=
{
	/// First parameter will be an integer from 1 to 10 named "Number of
	/// Iterations
	param_info(1, 100, "Number of Iterations", "Number of passes to apply"),
	
	/// Second parameter will be a double between 0 and 1
	param_info(2, 100, "Clusters", "Initial Number of Classes"),

	/// Second parameter will be a double between 0 and 1
	param_info(1.0, 10.0, "Fuzziness Index", "Accuracy of calculations")
};

/// Gives the probability of a gray level to belong to a class
double belonging(int k, histogram::gray_level_type level,
		centroid_set const& centroids,
		double fuzziness)
{
	bool const debug=false;

	double sum=0;
	double const dist_center_k_level=std::abs(double(level)-centroids[k]);
	double const pow=2/(fuzziness-1);

	for(unsigned  i=0; i<centroids.size(); ++i){
		double d=std::abs(double(level)-centroids[i]);
		if(debug)std::printf("Distance(): %f-%f=%f\n", double(level),
				centroids[i], d);
		if(std::abs(d)<0.0003){
			d=.5;
		}
		sum+=std::pow(dist_center_k_level/d, pow);
	}
	if(debug)std::printf("Belonging: %f\n", 1.0/sum);
	if(sum<0.0003) sum=.5;

	return 1.0/sum;
}


/// Actual implementation of the algorithm
void fcm_pass::operator() (histogram const& hist,
			centroid_set& centroids,
			progress_indicator* ind) const
{
	bool const debug=false;

	// First extract actual parameter values
	int const nr_passes=this->pass::parameter_value(0).int_value();
	int const nr_classes=this->pass::parameter_value(1).int_value();
	double const fuzziness=this->pass::parameter_value(2).double_value();

	if(centroids.empty()){
		// Initialize the centroids to the three most
		// common gray levels in the images
		std::map<int, histogram::gray_level_type> mp;
		for(histogram::const_iterator it=hist.begin(); it!=hist.end(); ++it){
			if(*it)
				mp.insert(std::make_pair(*it, it-hist.begin()));
		}
		
		std::map<int, histogram::gray_level_type>::iterator it=mp.begin();
		for(int i=0; i<nr_classes; ++i){
			if(it==mp.end()) break;
			centroids.push_back(it->second);
			++it;
		}
	}

	typedef std::pair<classification::iterator,
		classification::iterator> pair;

	int pass_count=nr_passes;
	classification c12n;
	std::vector<double> centroids_backup=centroids;
	while (pass_count--){
		classify(hist, centroids, &c12n);

		// For each class update the corresponding centroid
		for(unsigned i=0; i<centroids.size(); ++i){
			pair pr=c12n.equal_range(i);
			
			double weighted_sum=0; /// u(x)^m * x + u(x)^m *x + ..
			double weight_sum=0; /// u(x)^m + u(x)^m + ... 
			for(classification::iterator it=pr.first;
					it!=pr.second; ++it){
				double const weight=
					std::pow(belonging(i,
						it->second, centroids,
						fuzziness), fuzziness);
				if(debug)std::printf("*** weight: %f\n", weight);
				weighted_sum+=weight*double(it->second);
				weight_sum+=weight;
				if(debug)std::printf("weighted_sum:%f weight_sum:%f\n", 
					weighted_sum, weight_sum);
			}

			centroids_backup[i]=weighted_sum/weight_sum;
			if(debug)std::printf("CENTROID: %f\n", 
					weighted_sum/weight_sum);
		}
		centroids=centroids_backup;
	}
}

