/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef IMMUNE_SYSTEM_HPP
#define IMMUNE_SYSTEM_HPP

#include <functional>
#include <vector>
#include <cstdio>

/// Type traits generic/meaning less definition used by class immune_system
/// to retrieve information about the type of antibody it'll be working on
template<typename T>
struct antibody_traits
{
	/// Necessary typedef (should always be T)
	typedef T antibody_type;

	/// Integral type used to represent distance between antibodies
	typedef double distance_type;

	/// Mutates cl towards ag by the specified percent
	/// 1.0 means turn cl into ag, 0.0 means leave as is.
	static void mutate(antibody_type* cl, antibody_type ag,
			double percent);

	/// The distance (dissimilarity) between two antibodies
	static distance_type distance(antibody_type, antibody_type);

	/// Output an antibody to the specified stream (for debugging)
	void print(antibody_type, std::FILE*);
	
};

/// Immune systems hold a repertoire of antibodies that you can stimulate by
/// exposing it to antigens.
///
/// Antibody: The type used to represent antibodies (must have value semantics
///  as if they were to be stored in STL containers)
/// AntibodyTraits (defaults to antibody_traits<Antibody>) sets the functions
///  and types used to manipulate antibodies.
///
/// Basic usage looks like this:
/// - Decide which type of Antibody you'd like and write the corresponding
///    specialization of antibody_traits<>;
/// - Create an immune_system<my_antibody_type> is;
/// - Set up parameters like immune_system<>::zeta_percent,...
/// - Expose your immune system to antigens using
///    immune_system<>::stimulate(In,In)
/// - Retrieve the resulting antibody range using immune_system<>::begin()/end()
template<typename Antibody, typename AntibodyTraits = antibody_traits<Antibody> >
class immune_system
{
	typedef AntibodyTraits traits_type;
	
	typedef typename traits_type::antibody_type antibody_type;
	typedef typename traits_type::distance_type distance_type;
	typedef antibody_type       antigen_type;

	/// Given an antibody A returns a functor that can be used to compare two other
	/// antigens/antibodies for one which more compatible with A
	/// Sorting with this puts the more compatible antigen first.
	struct similarity_functor :
		std::binary_function<antigen_type, antigen_type, bool>
	{
		similarity_functor(antibody_type const& ab): m_template(ab) {}

		bool operator() (antigen_type const& a, antigen_type const& b) const
		{ return traits_type::distance(a, m_template) < traits_type::distance(b, m_template); }

		private:
		antibody_type m_template; 
	};

	/// Given an antibody A and a threshold T this unary predicate tells if the passed
	/// antibody/antigen is close enough to A (that is ||A-arg||<T)
	struct similarity_predicate:
		std::unary_function<antigen_type, bool>
	{
		similarity_predicate(antibody_type const& ab,
				distance_type threshold):
			m_antibody(ab), m_threshold(threshold){}

		bool operator() (antigen_type const& a) const
		{ return traits_type::distance(a, m_antibody) < m_threshold; }
		
		private:
		antibody_type m_antibody;
		distance_type m_threshold;
	};

	public:	
	/// Threshold for network suppression
	distance_type sigma_s;
	
	/// Threshold for clonal suppression
	distance_type sigma_d;

	/// Number of best responsive antibodies to select
	unsigned n;

	/// Will need this for reselecting a certain percent of best clones
	double zeta_percent;
	
	/// Type used for iterating over the internal range of antibodies
	typedef typename std::vector<antibody_type>::iterator iterator;

	/// Type used for iterating over the internal range of antibodies
	typedef typename std::vector<antibody_type>::const_iterator const_iterator;

	/// Create an immune system with an empty antibody repertoire
	immune_system(distance_type _sigma_s, distance_type _sigma_d,
			unsigned _n=3, double _zeta_percent=.3):
		sigma_s(_sigma_s), sigma_d(_sigma_d),
		n(_n), zeta_percent(_zeta_percent)
	{}

	// Artificially insert an antibody
	void insert(antibody_type const& a)
	{ m_repertoire.push_back(a); }

	/// Start of the current antibody repertoire
	iterator begin() { return m_repertoire.begin(); }
	
	/// Start of the current antibody repertoire
	const_iterator begin() const { return m_repertoire.begin(); }
	
	/// End of the current antibody repertoire
	iterator end() { return m_repertoire.end(); }
	/// End of the current antibody repertoire
	const_iterator end() const { return m_repertoire.end(); }

	/// Stimulates internal antibodies with a range of antigens
	template<class Input>
	void stimulate(Input beg, Input end);

	private:
	/// Mutates an antibody clone.
	/// \param cl The clone itself
	/// \param ag The antigen that has triggered the cloning
	void mutate(antibody_type* cl, antigen_type const& ag)
	{
		double const random=rand()/(RAND_MAX*3.0);
		distance_type const dist=traits_type::distance(*cl, ag);
		if(dist!=0){
			traits_type::mutate(cl, ag, 0.7 + random);
		}
	}

	/// Stimulates internal antibodies with a single antigen select most
	/// responsive antibodies of the repertoire, clone them, mutate them,
	/// reselect them and put the final clone set in result;
	///
	/// result is always cleared prior to action.
	void stimulate(antigen_type const& g, std::vector<antibody_type>& result);

	/// Tries to insert a new antibody into the repertoire, operation will
	/// not be effective if there is already another antibody with is
	/// similar up to a threshold
	void insert_antibody(antibody_type const& ab)
	{
		iterator it=std::find_if(m_repertoire.begin(),
				m_repertoire.end(), similarity_predicate(ab, sigma_s));
		if(it==m_repertoire.end()) m_repertoire.push_back(ab);
	}

	/// Current antibody repertoire
	std::vector<antibody_type> m_repertoire;
};

#endif // IMMUNE_SYSTEM_HPP
