// vim:set ft=cpp:
#ifndef IMMUNE_SYSTEM_TPP
#define IMMUNE_SYSTEM_TPP

#include "immune_system.hpp"


#include<cstdio>

#ifdef IMMUNE_DEBUG
/// Prints a labeled array of antibodies
template<typename Antibody>
void print(char const* label, std::vector<Antibody> const& v)
{
	typedef antibody_traits<Antibody> traits;
        std::printf("%s: ", label);
        for(typename std::vector<Antibody>::const_iterator it=v.begin();
                        it!=v.end(); ++it){
		traits::print(*it, stdout);
		std::printf(" ");
	}
        std::puts("");
}
#endif // IMMUNE_DEBUG

template<typename A, typename AT>
void immune_system<A,AT>::stimulate(typename immune_system<A,AT>::antigen_type const& g, 
		std::vector<antibody_type>& result)
{
#ifdef IMMUNE_DEBUG
	std::printf("****** Stimulating with: " );
	traits_type::print(g, stdout);
	puts("");
#endif

	// Start with a fresh result vector
	result.clear();

	iterator best_affinities=
		m_repertoire.begin()+std::min(this->n, m_repertoire.size());
		
	// Choose n of the most responsive antibodies in the repertoire
	if(this->n < m_repertoire.size()){
		std::nth_element(m_repertoire.begin(), 
				best_affinities, 
				m_repertoire.end(),
				similarity_functor(g));
	}
	
	// Clone then mutate the previously selected antibodies according to their
	// responsiveness
	for(iterator it=m_repertoire.begin(); 
			it!=best_affinities; ++it) {
		/// TODO: Should add a floor or ceil here
		distance_type const dist=traits_type::distance(*it, g);
		unsigned const nr_clones=dist?unsigned(m_repertoire.size()*
			(1.0/dist)+1):m_repertoire.size();

#ifdef IMMUNE_DEBUG
		std::printf("** Generating %d clones of ",
				nr_clones);
		traits_type::print(*it, stdout);
		std::puts("");
#endif

		for(unsigned i=0; i<nr_clones; ++i){
			antibody_type cl=*it;
			mutate(&cl, g);
			result.push_back(cl);
#ifdef IMMUNE_DEBUG
			std::printf("** Mutating ");
			traits_type::print(*it, stdout);
			std::printf(" ~~> ");
			traits_type::print(result.back(), stdout);
			std::puts("");
#endif
		}
	}

#ifdef IMMUNE_DEBUG
	::print<antibody_type>("initial clones", result);
#endif

	// Reselect best zeta percent clones
	best_affinities=result.begin()+(unsigned)(result.size()*this->zeta_percent);
	if(best_affinities!=result.begin()){
		std::nth_element(result.begin(),
				best_affinities,
				result.end(),
				similarity_functor(g));
		result.erase(best_affinities, result.end());
	}

#ifdef IMMUNE_DEBUG
	::print<antibody_type>("after zeta removal", result);
#endif

	// Remove the clones with affinity lower then sigma_d
	result.erase(std::remove_if(result.begin(), result.end(),
			std::not1(similarity_predicate(g, this->sigma_d))), result.end());
	
#ifdef IMMUNE_DEBUG
	::print<antibody_type>("after sigma_d removal", result);
#endif

	// Clonal suppression: eliminate duplicates (in the sens of the threshold
	// sigma_s)
	
	if(! result.empty()){
		for(unsigned i=0; i<result.size()-1; ++i)
			result.erase(std::remove_if(result.begin()+i+1,
						result.end(),
						similarity_predicate(result[i], sigma_s)),
					result.end());
	}

#ifdef IMMUNE_DEBUG
	::print<antibody_type>("after clonal suppression", result);	
#endif

}
// Stimulates internal antibodies with a range of antigens
template<typename A, typename AT>
template<typename Input>
void immune_system<A,AT>::stimulate(Input beg, Input end)
{ 
	std::vector<antibody_type> clones; 
	for(;beg!=end;++beg) {
		this->stimulate(*beg, clones);
		if(! clones.empty()){
			for(iterator i=clones.begin();
					i!=clones.end();
					++i){
				this->insert_antibody(*i);
			}
		} else {
			this->insert_antibody(*beg);
		}
#ifdef IMMUNE_DEBUG
		::print<antibody_type>("Antibody Repertoire", m_repertoire);
		puts("");
#endif
	}
}

#endif // IMMUNE_SYSTEM_TPP
