/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <iostream>
#include <set>

#include "jaccard_distance.hpp"
#include "pixel_region.hpp"

/// Jaccard distance between two specific classes in two images
///
/// image: Original (unmodified) image
/// ground_truth: Manual segmentation of 'image'
/// ground_class: Gray level of desired class in ground_truth
/// c12n: Result of segmentation applied on image
/// image_centroid: centroid of the desired class in image
double jaccard_distance_impl(GimpDrawable* image, GimpDrawable* ground_truth,
	guchar ground_class,
	classification const& c12n, 
	int image_centroid,
	int x, int y, int w, int h)
{
	// Travels through all the pixels (in x, y, w, h) of image
	pixel_region rgn_image(image, x, y, w, h, false, false);
	
	// Travels through all the pixels (in x, y, w, h) of ground_truth
	pixel_region rgn_truth(ground_truth, x, y, w, h, false, false);
	
	/// Number of pixels that have been classified in ground_class and with
	/// image_centroid
	int intersect_classes=0;
	
	/// Number of pixels that have been classified in ground_class or with
	/// image_centroid
	int union_classes=0;
		
	std::pair<classification::const_iterator,
				classification::const_iterator> pr=c12n.equal_range(image_centroid);

	// Will contain all gray levels in pr
	std::set<guchar> image_centroid_gray_levels;
	for(classification::const_iterator it=pr.first; it!=pr.second; ++it)
		image_centroid_gray_levels.insert(it->second);

	// Travels simultaneousy through rgn_image (with it) and through
	// rgn_truth (with jt)
	pixel_region::iterator it=rgn_image.begin();
	for(pixel_region::iterator jt=rgn_truth.begin(); jt!=rgn_truth.end();
		++jt){
		pixel_region::chunk const image_ch=*it;
		pixel_region::chunk const truth_ch=*jt;
		
		// Travels simultaneously through image_ch (with i_ch) and
		// through truth_ch (with t_ch) 
		guchar const* i_ch=image_ch.begin;
		for(guchar const* t_ch=truth_ch.begin; t_ch!=truth_ch.end; 
				t_ch+=ground_truth->bpp){

			bool belongs_to_image_centroid=false;

			// look for *i_ch in image_centroid_gray_levels
			if(image_centroid_gray_levels.find(*i_ch)!=image_centroid_gray_levels.end()) {
				belongs_to_image_centroid=true; 
			}				
			
			bool belongs_to_truth_class=false;
			if(ground_class==*t_ch)
				belongs_to_truth_class=true;

			if(belongs_to_truth_class || belongs_to_image_centroid) union_classes++;
			if(belongs_to_truth_class && belongs_to_image_centroid) intersect_classes++;
			
			i_ch+=image->bpp;
		}		
		
		++it;
	}
	if(false)std::cout<<"double(intersect_classes) "<<double(intersect_classes)<<
		" double(union_classes) "<<double(union_classes)<<std::endl;
	return double(intersect_classes)/double(union_classes);
	
}
	
void jaccard_distance(GimpDrawable* image, 
	classification const& image_c12n,
	centroid_set const& image_centers,
	GimpDrawable* ground_truth,
	histogram const& ground_hist,
	int x, int y, int w, int h,
	std::vector<std::pair<int, double> >* out)
{
	out->clear();

	// For each possible gray level (class) in the manual segmentation
	for(histogram::const_iterator it=ground_hist.begin(); 
		it!=ground_hist.end(); ++it){
		if(*it==0) continue; // consider only possible gray levels

		std::pair<int, double> pair; pair.first=0; pair.second=0;
		
		// For each possible centroid (class) in the automatic
		// segmentation
		for(unsigned jt=0; jt!=image_centers.size(); ++jt){
			double const index=jaccard_distance_impl(image, ground_truth,
				it-ground_hist.begin(), image_c12n, jt, x, y, w, h);
			
			// Keep only the best match (highest jaccard index)
			if(index > pair.second) {
				pair.second=index;
				pair.first=jt;
			}					
		}
		out->push_back(pair);
	}
}


