/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <vector>
#include <cstdlib>
#include <sstream>
#include <cassert>
#include <iterator>
#include <cstring>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <gtk/gtk.h>
#include <cmath>

#include "centroid_set.hpp"
#include "compute_histogram.hpp"
#include "progress_indicator.hpp"
#include "ainet_pass.hpp"
#include "kmeans_pass.hpp"
#include "multipass_dialog.hpp"
#include "rgb_to_gray.hpp"
#include "pixel_class.hpp"
#include "register_passes.hpp"
#include "result_dialog.hpp"
#include "register_fitness_functions.hpp"
#include "register_pixel_classifiers.hpp"
#include "colors.hpp"

/// An interactive-only plugin that allows to stack passes to dynamically
/// create the algorithm you want before appying it

// Description of parameters the plug-in will take.
// 
// For a filter type plug-in, this includes, (at least), the run mode
// (interactive or not), the input image and the drawable to work on.
GimpParamDef param_def[]=
{
	{ GIMP_PDB_INT32, (gchar*)"run-mode", (gchar*)"Run mode"},
	{ GIMP_PDB_IMAGE, (gchar*)"image", (gchar*)"Input Image" },
	{ GIMP_PDB_DRAWABLE, (gchar*)"drawable", (gchar*)"Input drawable"}
};

// Description of the return values.
//
// Must at least include a status code.
GimpParamDef return_values_def[]=
{
	{ GIMP_PDB_STATUS, (gchar*)"status", (gchar*)"Status code" }
};

// Registers a new procedure in the PDB, and setup a menu item to trigger it.
void query()
{
	// Register plug-in-hello in the PDB
	gimp_install_procedure(
			"plug-in-multipass",
			"Segment a gray scale image by chaining algorithms.",
			"Segment a gray scale image by chaining algorithms.",
			"Anis Benyelloul",
			"2007",
			"03/11/2007",
			"_Multipass Segmentation",
			"GRAY* RGB*",
			GIMP_PLUGIN,
			sizeof(::param_def)/sizeof(::param_def[0]),
			sizeof(return_values_def)/sizeof(return_values_def[0]),
			::param_def,
			::return_values_def);

	// Setup an appropriate menu item
	gimp_plugin_menu_register("plug-in-multipass",
			"<Image>/Segmentation");
}

namespace
{
/// Variable size structure to lay out info about pass (index + params)
struct pass_layout
{
	int index;
	param parameters[1];
};

/// Index of the given pass' label in the repertoire
int pass_index(char const* label)
{
	pass_repertoire* rep=pass_repertoire::instance();
	int i=0;
	for(pass_repertoire::iterator it=rep->begin();
			it!=rep->end(); ++it){
		if(std::strcmp((*it)->label(),label)==0) {
			return i;
		} else ++i;
	}
	assert(false);
	return 0;
}

/// Clones and returns a pass from the repertoire given its index
pass* get_pass_by_index(int index)
{
	pass_repertoire* rep=pass_repertoire::instance();
	pass_repertoire::iterator it=rep->begin();
	if(index)std::advance(it, index);

	return (*it)->clone();
}

/// Serializes an algorithm into raw bytes sequence so that it can
/// be stored then broght back with deserialize_algo()
void serialize_algo(std::vector<pass*> const& algo,
		std::vector<unsigned char>& output)
{
	output.clear();
	for(std::vector<pass*>::const_iterator it=algo.begin();
			it!=algo.end(); ++it){
		int const index=pass_index((*it)->label());
		int const nr_params=(*it)->nr_parameters();

		unsigned const old_size=output.size();
		output.resize(output.size()+sizeof(int)+sizeof(param)*nr_params);

		pass_layout* const pl=reinterpret_cast<pass_layout*>(&output[old_size]); 
		pl->index=index;
		for(int i=0; i<nr_params; ++i){
			pl->parameters[i]=(*it)->parameter_value(i);
		}
	}
}

/// Deserialize an algo out of a byte sequence created by serialize_algo
void deserialize_algo(std::vector<pass*>& algo,
		std::vector<unsigned char> const& input)
{
	std::vector<unsigned char>::const_iterator it=input.begin();
	algo.clear();

	while(it<input.end()) {
		pass_layout const* const pl=reinterpret_cast<pass_layout const*>(&*it);
		pass* p=get_pass_by_index(pl->index);
		for(unsigned i=0; i<p->nr_parameters(); ++i)
			p->parameter_value(i)=pl->parameters[i];
		
		it+=sizeof(int)+sizeof(param)*p->nr_parameters();
		algo.push_back(p);
	}
}

/// Sets the the value(0-255) (HSV) of an rgb triplet
void rgb_set_value(rgb * color, gint value)
{
	gint h=color->r, s=color->g, v=color->b;
	gimp_rgb_to_hsv_int(&h, &s, &v);
	gimp_hsv_to_rgb_int(&h, &s, &value);
	color->r=h; color->g=s; color->b=value;
}

/// Given a drawable one of its tiles and a pointer inside the tile should
/// return the X, Y position of the pixel in the drawable
std::pair<gint, gint> memory_to_xy(GimpDrawable* drawable,
		GimpTile* tile,
		guchar* ptr)
{
	guint const tilenum=tile->tile_num;
	guint const tile_col=tilenum%drawable->ntile_cols;
	guint const tile_row=tilenum/drawable->ntile_cols;
	
	/// Pos of upper left corner of tile
	gint const tile_x=tile_col*gimp_tile_width();
	gint const tile_y=tile_row*gimp_tile_height();

	/// Pos of ptr relative to tile upper left
	std::ptrdiff_t disp=(ptr-tile->data)/drawable->bpp;
	gint const x=disp%tile->ewidth;
	gint const y=disp/tile->ewidth;
	
	return std::make_pair(x+tile_x, y+tile_y);
}

} // anonymous namespace

/// Creates an image that visually represent the result
/// of computation by assigning a color to each class
void create_result_image(
		char const* title,
		gint32 drawable_id,
		centroid_set const& centroids,
		histogram const& hist,
		pixel_classifier* pixel_cls)
{


	GimpDrawable* drawable=gimp_drawable_get(drawable_id);
	
	gint32 image_id=gimp_image_new(drawable->width,
			drawable->height, GIMP_RGB);	
	gint32 new_layer_id=gimp_layer_new(image_id, title,
			drawable->width, 
			drawable->height,
			GIMP_RGB_IMAGE,
			100.0, GIMP_NORMAL_MODE);
	GimpDrawable* new_drawable=gimp_drawable_get(new_layer_id);
	gimp_image_add_layer(image_id, new_layer_id, 1);
	gimp_drawable_fill(new_layer_id, GIMP_BACKGROUND_FILL);
	gimp_display_new(image_id);


        {
		classification c12n; classify(hist, centroids, &c12n);

		GimpPixelRgn gimp_rgn[1];
		gimp_pixel_rgn_init(gimp_rgn, drawable, 0, 0, 
				drawable->width, drawable->height,
				false, false);

		gimp_tile_cache_ntiles(10);
                pixel_region in_rgn(drawable, 0, 0, drawable->width,
                                drawable->height, false, false);
                pixel_region out_rgn(new_drawable, 0, 0, new_drawable->width,
                                new_drawable->height, true, true);
                pixel_region::iterator out_it=out_rgn.begin();
                for(pixel_region::iterator in_it=in_rgn.begin();
                                in_it!=in_rgn.end(); ++in_it){
                        guchar* out=out_it->begin;
                        for(guchar* ch=in_it->begin; ch!=in_it->end; ch+=drawable->bpp){
                                rgb color;
				std::pair<gint, gint> pos=memory_to_xy(drawable,
						in_it.tile(),
						ch);
				int const pclass=pixel_cls->operator()(hist,
						centroids, c12n,
						drawable, 
						gimp_rgn,
						*ch,
						pos.first, pos.second, 0.5);
				//gint const sat=(gint)std::ceil(255-std::abs(double(*ch)-centroids[pclass]));
                                color=::colors[pclass];
                                //rgb_set_value(&color, sat);

                                out[0]=color.r;
                                out[1]=color.g;
                                out[2]=color.b;
                                out+=new_drawable->bpp;
                        }
                        assert(out==out_it->end);
                        out_it++;
                }
                assert(out_it==out_rgn.end());
        }


	gimp_drawable_flush(new_drawable);
	gimp_drawable_merge_shadow(new_layer_id, true);
	gimp_drawable_update(new_layer_id, 0, 0, new_drawable->width, new_drawable->height);


	gimp_drawable_detach(drawable);
	gimp_drawable_detach(new_drawable);
}

/// Calls gimp_progress_update()
class gimp_progress_indicator : public progress_indicator
{
	public:
	virtual void operator() (double percent)
	{ gimp_progress_update(percent); }
};

#if 0
void dump_colors()
{
	gint r, g, b;

	gint sats[]={255, 191, 127, 63, 31, 95, 159, 223, 15, 
		47, 79, 111, 143, 175, 212, 239, 7 };

	for (int i=0; i<sizeof(sats)/sizeof(sats[0]); ++i){
		for(int c=0; c<17; ++c){
			r=::colors[c].r;
			g=::colors[c].g;
			b=::colors[c].b;

			gimp_rgb_to_hsv_int(&r, &g, &b);
			b=sats[i];
			gimp_hsv_to_rgb_int(&r, &g, &b);

			std::printf("{%d, %d, %d}, \n", r, g, b);
		}
	}


}
#endif // #if 0

void run(gchar const* name,
		gint nparams,

		GimpParam const* params,
		gint* nr_return_values,
		GimpParam ** return_values)
{

//	std::random_shuffle(::colors, ::colors+50);

	// Setup mandatory return value facilities
	static GimpParam ret;
	ret.type=GIMP_PDB_STATUS;
	ret.data.d_status=GIMP_PDB_SUCCESS; // set return value to SUCCESS for now.
	*nr_return_values=1;
	*return_values=&ret;

	// Retrieve passed parameters
	gint32 const run_mode=params[0].data.d_int32;
	//gint32 const image_id=params[1].data.d_image;
	GimpDrawable* drawable=gimp_drawable_get(params[2].data.d_drawable);

	::register_passes();
	::register_fitness_functions();
	::register_pixel_classifiers();
	// Do the job
	if(run_mode==GIMP_RUN_INTERACTIVE)
	{
		gimp_ui_init("plug-in-multipass", false);
		
		char const* const gimp_data_id="plug-in-multipass";
		std::vector<pass*> algo; 
		pixel_classifier* pixel_cls;
		int const bytes=gimp_get_data_size(gimp_data_id);
		if(bytes!=0){
			std::vector<unsigned char> buffer(bytes);
			gimp_get_data(gimp_data_id, &buffer[0]);
			deserialize_algo(algo, buffer);
		} else {
			algo.push_back(new ainet_pass());
		}

		if(multipass_dialog(drawable, algo, &pixel_cls)){
			
			centroid_set centroids(0);
			gimp_progress_indicator ind;
			histogram hist;

			::compute_histogram(drawable, &hist, 0, 0, drawable->width, drawable->height);
			
			std::vector<centroid_set> results;
			
			// Apply all passes in turn
			for(std::vector<pass*>::iterator it=algo.begin();
					it!=algo.end(); ++it){
				gimp_progress_init((*it)->label());
				gimp_progress_update(0.0);
				(*it)->operator() (hist, centroids, &ind);
				results.push_back(centroids);
				gimp_progress_update(1.0);
			}
			gimp_progress_init("");
			gimp_progress_update(0.0);

			std::ostringstream oss;
			for(std::vector<pass*>::iterator it=algo.begin();
					it!=algo.end(); ++it){
				oss<<(*it)->label()<<" ";
			}
			create_result_image(oss.str().c_str(),
					drawable->drawable_id,
					centroids, hist, pixel_cls);

			std::vector<unsigned char> buffer;
			serialize_algo(algo, buffer);
			gimp_set_data(gimp_data_id, &buffer[0], buffer.size());

			// Cleanup
			gimp_displays_flush();
			
			result_dialog(gimp_drawable_get_image(drawable->drawable_id),
					algo, hist, results);
		} // if(multipass_dialog)
	}

	gimp_drawable_detach(drawable);
}

GimpPlugInInfo PLUG_IN_INFO={
	NULL,
	NULL,
	query,
	run
};

MAIN()
