/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <vector>
#include <cstdlib>
#include <cerrno>
#include <cstring>
#include <cassert>
#include <iterator>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <gtk/gtk.h>

#include "classification.hpp"
#include "fitness_function_repertoire.hpp"
#include "centroid_set.hpp"
#include "compute_histogram.hpp"
#include "progress_indicator.hpp"
#include "ainet_pass.hpp"
#include "kmeans_pass.hpp"
#include "multipass_batch_dialog.hpp"
#include "pixel_class.hpp"
#include "register_passes.hpp"
#include "register_fitness_functions.hpp"
#include "load_manual_segmentation.hpp"
#include "jaccard_distance.hpp"

/// An interactive-only plugin that allows to stack passes to dynamically
/// create the algorithm you want before appying it

// Description of parameters the plug-in will take.
// 
// For a filter type plug-in, this includes, (at least), the run mode
// (interactive or not), the input image and the drawable to work on.
GimpParamDef param_def[]=
{
	{ GIMP_PDB_INT32, (gchar*)"run-mode", (gchar*)"Run mode"}
};

// Description of the return values.
//
// Must at least include a status code.
GimpParamDef return_values_def[]=
{
	{ GIMP_PDB_STATUS, (gchar*)"status", (gchar*)"Status code" }
};

/// Given the results from jaccard_distance (see jaccard_distance.hpp) returns
/// the mean of all jaccard indexes  
double jaccard_mean(std::vector< std::pair<int, double> > const& res)
{
	double mean=0.0;
	for(std::vector<std::pair<int,double> >::const_iterator it=res.begin();
			it!=res.end(); ++it)
		mean+=it->second;
	return mean/res.size();
}

// Registers a new procedure in the PDB, and setup a menu item to trigger it.
void query()
{
	// Register plug-in-multipass-batch in the PDB
	gimp_install_procedure(
			"plug-in-multipass-batch",
			"Segment a gray scale image by chaining algorithms.",
			"Segment a gray scale image by chaining algorithms.",
			"Anis Benyelloul",
			"2007",
			"03/11/2007",
			"_Multipass Segmentation (Batch Mode)",
			"",
			GIMP_PLUGIN,
			sizeof(::param_def)/sizeof(::param_def[0]),
			sizeof(return_values_def)/sizeof(return_values_def[0]),
			::param_def,
			::return_values_def);

	// Setup an appropriate menu item
	gimp_plugin_menu_register("plug-in-multipass-batch",
			"<Toolbox>/Segmentation");
}

namespace
{
/// Variable size structure to lay out info about pass (index + params)
struct pass_layout
{
	int index;
	param parameters[1];
};

/// Index of the given pass' label in the repertoire
int pass_index(char const* label)
{
	pass_repertoire* rep=pass_repertoire::instance();
	int i=0;
	for(pass_repertoire::iterator it=rep->begin();
			it!=rep->end(); ++it){
		if(std::strcmp((*it)->label(),label)==0) {
			return i;
		} else ++i;
	}
	assert(false);
	return 0;
}

/// Clones and returns a pass from the repertoire given its index
pass* get_pass_by_index(int index)
{
	pass_repertoire* rep=pass_repertoire::instance();
	pass_repertoire::iterator it=rep->begin();
	if(index)std::advance(it, index);

	return (*it)->clone();
}

/// Serializes an algorithm into raw bytes sequence so that it can
/// be stored then brought back with deserialize_algo()
void serialize_algo(std::vector<pass*> const& algo,
		std::vector<unsigned char>& output)
{
	output.clear();
	for(std::vector<pass*>::const_iterator it=algo.begin();
			it!=algo.end(); ++it){
		int const index=pass_index((*it)->label());
		int const nr_params=(*it)->nr_parameters();

		unsigned const old_size=output.size();
		output.resize(output.size()+sizeof(int)+sizeof(param)*nr_params);

		pass_layout* const pl=reinterpret_cast<pass_layout*>(&output[old_size]); 
		pl->index=index;
		for(int i=0; i<nr_params; ++i){
			pl->parameters[i]=(*it)->parameter_value(i);
		}
	}
}

/// Deserialize an algo out of a byte sequence created by serialize_algo
void deserialize_algo(std::vector<pass*>& algo,
		std::vector<unsigned char> const& input)
{
	std::vector<unsigned char>::const_iterator it=input.begin();
	algo.clear();

	while(it<input.end()) {
		pass_layout const* const pl=reinterpret_cast<pass_layout const*>(&*it);
		pass* p=get_pass_by_index(pl->index);
		for(unsigned i=0; i<p->nr_parameters(); ++i)
			p->parameter_value(i)=pl->parameters[i];
		
		it+=sizeof(int)+sizeof(param)*p->nr_parameters();
		algo.push_back(p);
	}
}


} // anonymous namespace

/// Progress bar for the overall activity
GtkWidget* overall_progress;

/// Progress bar for current algorithm activity
GtkWidget* algo_progress;

/// Updates a given double
class double_progress_indicator : public progress_indicator
{
	public:
		double_progress_indicator(double* progress): m_progress(progress) {}
	
		virtual void operator() (double percent)
		{ *m_progress=percent; }
	
	private:
	double* m_progress;
};

// Creates and returns a GtkDialog with two progress bars and a cancel
// button to show the progression of operation (both overall and per algorithm)
GtkWidget* create_progress_dialog()
{
	GtkWidget* dialog=gtk_dialog_new_with_buttons("Progress", NULL, GtkDialogFlags(0),
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
	
	GtkWidget* overall_frame=gtk_frame_new("Overall Progression");
	GtkWidget* algo_frame=gtk_frame_new("Algorithm Progression");

	::overall_progress=gtk_progress_bar_new();
	::algo_progress=gtk_progress_bar_new();

	gtk_container_add(GTK_CONTAINER(overall_frame), overall_progress);
	gtk_container_add(GTK_CONTAINER(algo_frame), algo_progress);
	gtk_widget_show(overall_progress);
	gtk_widget_show(algo_progress);
	
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), overall_frame, false, false, 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), algo_frame, false, false, 10);
	gtk_widget_show(algo_frame);
	gtk_widget_show(overall_frame);

	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(overall_progress), 0.5);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(overall_progress), "50%");
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(algo_progress), 0.7);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(algo_progress), "70%");
	
	gtk_window_set_default_size(GTK_WINDOW(dialog), 400, 100);
	return dialog;
}

/// The benchmark batch() will be acting on
std::vector<std::string>* benchmark;

/// Where batch() should send results
FILE* bench_result;

/// The algorithm bench() should apply
std::vector<pass*>* bench_algo;

/// Accessed by batch and timeout_handler
double overall_progress_fraction=0.0;

/// Accessed by batch and timeout_handler
double algo_progress_fraction=0.0;

/// Accessed by batch and timeout_handler
char const* algo_progress_label="";

/// Flag chacked by timeout_handler to know when to stop
bool timeout_handler_stop=false;

gboolean timeout_handler(gpointer)
{
	if(::timeout_handler_stop) return false;

	char buffer[100];
	std::sprintf(buffer, "%.0f%%", ::overall_progress_fraction*100);
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(::overall_progress),
			::overall_progress_fraction);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(::overall_progress),
			buffer);
	std::sprintf(buffer, "%s (%.0f%%)", ::algo_progress_label,
			::algo_progress_fraction*100);
	

	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(::algo_progress),
			::algo_progress_fraction);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(::algo_progress),
			buffer);
	std::printf(".");
	return true;

}

/// Apply bench_algo to benchmark and sends results in bench_result
void batch(gpointer data)
{
	bool const debug=false;

	GtkDialog* dialog=GTK_DIALOG(data);
	::overall_progress_fraction=0.0;
	::algo_progress_fraction=0.0;
	
	double_progress_indicator ind(&::algo_progress_fraction);

	fitness_function_repertoire& rep=*fitness_function_repertoire::instance();
	histogram hist;
	classification c12n;
	centroid_set centroids;
	
	std::fprintf(::bench_result, "\n\nAlgorithm: ");
	for(std::vector<pass*>::iterator it=::bench_algo->begin();
			it!=::bench_algo->end(); ++it){
		std::fprintf(::bench_result, "%s ",
				(*it)->label());
	}
	std::fprintf(::bench_result, "\n");
	
	std::fprintf(::bench_result, "Image\tNumber of classes\tExecuction Time");
	for(fitness_function_repertoire::iterator it=rep.begin(); it!=rep.end(); ++it)
		std::fprintf(::bench_result, "\t%s", (*it)->label());
	
	std::fprintf(::bench_result, "\tJaccard Distance (Supervised) ");
	std::fprintf(::bench_result, "\n");


	GTimer* timer=g_timer_new();
	g_timer_stop(timer);
	
	for(std::vector<std::string>::iterator it=::benchmark->begin();
			it!=::benchmark->end(); ++it){
		::overall_progress_fraction=(it-::benchmark->begin())/double(::benchmark->size());

		gint32 image_id=gimp_file_load(GIMP_RUN_NONINTERACTIVE,
				it->c_str(), it->c_str());
		if(image_id<0) continue;
		else std::printf("Got %s\n", it->c_str());

		GimpDrawable* ground_truth_drawable=load_manual_segmentation(image_id);
		histogram ground_truth_hist;

		if(ground_truth_drawable){
			compute_histogram(ground_truth_drawable, &ground_truth_hist,
					0,0,ground_truth_drawable->width,
					ground_truth_drawable->height);
		}

		gint nr_layers;
		gint* layers=gimp_image_get_layers(image_id, &nr_layers);
		if(nr_layers==0) continue;

		GimpDrawable* drawable=gimp_drawable_get(layers[0]);

		compute_histogram(drawable, &hist, 0, 0, drawable->width, drawable->height);
		if(debug)for(histogram::iterator jt=hist.begin(); jt!=hist.end(); ++jt)
			std::printf("hist[%d]=%d\n", jt-hist.begin(), *jt);

		centroids.clear();

		g_timer_start(timer);
		for(std::vector<pass*>::iterator jt=::bench_algo->begin(); jt!=::bench_algo->end(); ++jt){
			::algo_progress_label=(*jt)->label();
			::algo_progress_fraction=0.0;
			(*jt)->operator () (hist, centroids, &ind);
			::algo_progress_fraction=1.0;
		}
		g_timer_stop(timer);

		if(debug){
			std::printf("On %s ", it->c_str());
			for(centroid_set::iterator jt=centroids.begin(); jt!=centroids.end(); ++jt)
				std::printf("%f ", *jt);
			std::printf("\n");
		}

		classify(hist, centroids, &c12n);
		std::fprintf(::bench_result, "%d\t%d\t%.2f", it-::benchmark->begin(), centroids.size(),
				g_timer_elapsed(timer, NULL));
		for(fitness_function_repertoire::iterator it=rep.begin(); it!=rep.end(); ++it){
			std::fprintf(::bench_result, "\t%.2f", (*it)->operator() (hist, centroids, c12n));

		}
		if(ground_truth_drawable)
		{
			std::vector< std::pair<int,double> > res;
			jaccard_distance(drawable, c12n, centroids, ground_truth_drawable,
					ground_truth_hist,
					0, 0, drawable->width, drawable->height,
					&res);
			std::fprintf(::bench_result, "\t%.2f",
					jaccard_mean(res));

		}
		std::fprintf(::bench_result, "\n");

		gimp_drawable_detach(drawable);
	}
	fclose(::bench_result);
	g_timer_destroy(timer);
	gtk_dialog_response(dialog, 1);
}

void run(gchar const* name,
		gint nparams,

		GimpParam const* params,
		gint* nr_return_values,
		GimpParam ** return_values)
{
	// Setup mandatory return value facilities
	static GimpParam ret;
	ret.type=GIMP_PDB_STATUS;
	ret.data.d_status=GIMP_PDB_SUCCESS; // set return value to SUCCESS for now.
	*nr_return_values=1;
	*return_values=&ret;

	// Retrieve passed parameters
	gint32 const run_mode=params[0].data.d_int32;

	::register_passes();
	::register_fitness_functions();

	// Do the job
	if(run_mode==GIMP_RUN_INTERACTIVE)
	{
		gimp_ui_init("plug-in-multipass-batch", false);
		g_thread_init(NULL);
		
		char const* const gimp_data_id="plug-in-multipass-batch";
		std::vector<pass*> algo; 
		::bench_algo=&algo;

		int const bytes=gimp_get_data_size(gimp_data_id);
		if(bytes!=0){
			std::vector<unsigned char> buffer(bytes);
			gimp_get_data(gimp_data_id, &buffer[0]);
			deserialize_algo(algo, buffer);
		} else {
			algo.push_back(new ainet_pass());
		}

		std::vector<std::string> benchmark;
		::benchmark=&benchmark;

		std::string result_filename;
		if (multipass_batch_dialog(algo, benchmark, result_filename))
		{
			for(std::vector<std::string>::iterator it=benchmark.begin();
					it!=benchmark.end(); ++it){
				std::printf("BENCH: %s\n", it->c_str());
			}
			std::printf("OUT =>%s\n", result_filename.c_str());

			::bench_result=std::fopen(result_filename.c_str(), "a");
			if(!::bench_result){
				GtkWidget* dialog=gtk_message_dialog_new(NULL,
						GtkDialogFlags(0),
						GTK_MESSAGE_ERROR,
						GTK_BUTTONS_OK,
						"Could not open '%s': %s",
						result_filename.c_str(),
						std::strerror(errno));
				gtk_dialog_run(GTK_DIALOG(dialog));
				gtk_widget_destroy(dialog);
				return;
			}
			GtkWidget* progress_dialog=create_progress_dialog();
			g_thread_create(GThreadFunc(batch),
					progress_dialog, false, NULL);
			g_timeout_add(100, timeout_handler, NULL);
			int res=gtk_dialog_run(GTK_DIALOG(progress_dialog));
			::timeout_handler_stop=true;
			gtk_widget_destroy(progress_dialog);
			if(res>0){
				GtkWidget* dialog=gtk_message_dialog_new(NULL,
						GtkDialogFlags(0),
						GTK_MESSAGE_INFO,
						GTK_BUTTONS_OK,
						"Operation Successful");
				gtk_dialog_run(GTK_DIALOG(dialog));
				gtk_widget_destroy(dialog);
				return;
			
			}
		}
	}

}

GimpPlugInInfo PLUG_IN_INFO={
	NULL,
	NULL,
	query,
	run
};

MAIN()
