/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <iterator>
#include <map>
#include <cstdio>
#include <cerrno>
#include <cstring>
#include <sstream>

#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "multipass_dialog.hpp"
#include "pixel_class.hpp"
#include "ui_build.hpp"
#include "pixel_region.hpp"
#include "pass_repertoire.hpp"
#include "compute_histogram.hpp"
#include "progress_indicator.hpp"
#include "fitness_function_repertoire.hpp"

namespace
{
void remove_button_handler(GtkWidget* widget, gpointer data);
void up_button_handler(GtkWidget* button, gpointer data);
void down_button_handler(GtkWidget* button, gpointer data);

// Names the columns of the benchmark file list
enum {column_icon, column_filename, nr_column };

// The model (MVC) of the benchmark file list
GtkListStore* list_store;

// The view (MVC) of the benchmark file list;
GtkWidget* tree_view;

// The GimpFileEntry receiving result storage location
GtkWidget*  result_file_entry;

// The current benchmark filename being edited (or empty)
std::string* benchmark_filename;

// The GimpDialog created in multipass_batch_dialog
GtkWidget* gimp_dialog;

// Matches an entry panel (as returned by ui_build)
// to the pass uppon which it is acting
std::map<GtkWidget*, pass*>* panel_pass;

// The vbox containing a panel per pass
GtkWidget* pass_panel_vbox;

// handler for "value-changed" of slides and scales
void value_changed_handler(GtkWidget* value, gpointer)
{
}

// handler for the combo box for selecting the pass type
void pass_combo_box_handler(GtkWidget* combo, gpointer data)
{
	GtkWidget* parent=GTK_WIDGET(data);
	GtkWidget* pass_ui=GTK_BIN(parent)->child;

	int active=gtk_combo_box_get_active(GTK_COMBO_BOX(combo));

	if(active<0)active=0;

	pass_repertoire::iterator it=pass_repertoire::instance()->begin();
	std::advance(it, active);

	pass* new_pass=(*it)->clone();
	GtkWidget* new_pass_ui=ui_build(new_pass, value_changed_handler);
	
	std::map<GtkWidget*, pass*>::iterator jt=panel_pass->find(
			gtk_widget_get_parent(gtk_widget_get_parent(pass_ui)));
	if(jt!=panel_pass->end()){
		delete jt->second;
		jt->second=new_pass;
	}

	gtk_widget_destroy(pass_ui);

	gtk_container_add(GTK_CONTAINER(parent), new_pass_ui);
	gtk_widget_show(new_pass_ui);
}

// Creates and returns a combo_box that contains all available pass' labels
// initial selection is set to passed string if possible
GtkWidget* label_chooser(char const* init, GtkWidget* pass_param_parent)
{
	pass_repertoire& rep=*pass_repertoire::instance();
	GtkWidget* combo=gtk_combo_box_new_text();

	gint index=-1;
	int i=0;
	for(pass_repertoire::iterator it=rep.begin();
			it!=rep.end(); ++it, ++i){
		if(std::strcmp((*it)->label(), init)==0)
			index=i;
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo),
				(*it)->label());
	}
	if(index>=0)
		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), index);
	else 
		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
	g_signal_connect(G_OBJECT(combo), "changed",
			G_CALLBACK(pass_combo_box_handler),
			pass_param_parent);
	return combo;
}

// Creates and returns a panel for a pass
GtkWidget* pass_panel(pass* p)
{
	GtkWidget* frame=gtk_frame_new("");
	
	GtkWidget* pass_param=ui_build(p, value_changed_handler);
	panel_pass->insert(std::make_pair(frame, p));

	GtkWidget* exp=gtk_expander_new("Parameters");
	
	GtkWidget* title_bar=gtk_hbox_new(false, 20);
	// Fill the title_bar
	{
		GtkWidget* label_ch=label_chooser(p->label(), exp);
		gtk_box_pack_start(GTK_BOX(title_bar),
				label_ch, false, false, 0);
		GtkWidget* button=gtk_button_new_from_stock(GTK_STOCK_REMOVE);
		GtkWidget* up_button=gtk_button_new_from_stock(GTK_STOCK_GO_UP);
		GtkWidget* down_button=gtk_button_new_from_stock(GTK_STOCK_GO_DOWN);
		g_signal_connect(G_OBJECT(button), "clicked", 
				G_CALLBACK(remove_button_handler),
				frame);
		g_signal_connect(G_OBJECT(up_button),
				"clicked",
				G_CALLBACK(up_button_handler),
				frame);
		g_signal_connect(G_OBJECT(down_button),
				"clicked",
				G_CALLBACK(down_button_handler),
				frame);
		gtk_box_pack_end(GTK_BOX(title_bar),
				button, false, false, 0);
		gtk_box_pack_end(GTK_BOX(title_bar),
				down_button, false, false, 0);
		gtk_box_pack_end(GTK_BOX(title_bar),
				up_button, false, false, 0);
		gtk_widget_show(label_ch);
		gtk_widget_show(button);
		gtk_widget_show(up_button);
		gtk_widget_show(down_button);
	}

	gtk_container_set_border_width(GTK_CONTAINER(title_bar), 10);

	gtk_frame_set_label_widget(GTK_FRAME(frame), title_bar);
	gtk_widget_show(title_bar);

	gtk_container_add(GTK_CONTAINER(exp), pass_param);
	gtk_widget_show(pass_param);

	gtk_container_set_border_width(GTK_CONTAINER(exp), 5);
	gtk_container_add(GTK_CONTAINER(frame), exp);
	gtk_widget_show(exp);
	gtk_widget_show(frame);
	
	return frame;
}

// Handler of the "+ Add" button
void add_button_handler(GtkWidget* widget, gpointer data)
{
	GtkWidget* vbox=GTK_WIDGET(data);
	
	pass* p=*pass_repertoire::instance()->begin();
	GtkWidget* panel=pass_panel(p->clone());
	gtk_box_pack_start(GTK_BOX(vbox), panel, 0, 0, false);
	gtk_widget_show(panel);
}

// Handler of the "- Remove button"
void remove_button_handler(GtkWidget* widget, gpointer data)
{
	GtkWidget* panel=GTK_WIDGET(data);
	GtkWidget* parent=gtk_widget_get_parent(panel);
	
	if(::panel_pass->size()==1) return;

	std::map<GtkWidget*, pass*>::iterator it=::panel_pass->find(panel);
	if(it!=panel_pass->end()){
		delete it->second;
		panel_pass->erase(it);
	}
	gtk_container_remove(GTK_CONTAINER(parent),
			panel);
}

// Handler of the Up button
void up_button_handler(GtkWidget* button, gpointer data)
{
	GtkWidget* panel=GTK_WIDGET(data);
	GtkWidget* box=gtk_widget_get_parent(panel);

	GValue pos;
	std::fill((char*)&pos, (char*)(&pos+1), 0);

	g_value_init(&pos, G_TYPE_INT);

	gtk_container_child_get_property(GTK_CONTAINER(box),
			panel, "position", &pos);

	int posint=g_value_get_int(&pos);

	if(posint>0){
		gtk_box_reorder_child(GTK_BOX(box), panel,
				posint-1);
	}
}

// Handler of the down button
void down_button_handler(GtkWidget* button, gpointer data)
{
	GtkWidget* panel=GTK_WIDGET(data);
	GtkWidget* box=gtk_widget_get_parent(panel);

	GValue pos;
	std::fill((char*)&pos, (char*)(&pos+1), 0);

	g_value_init(&pos, G_TYPE_INT);

	gtk_container_child_get_property(GTK_CONTAINER(box),
			panel, "position", &pos);

	int posint=g_value_get_int(&pos);
	
	if(unsigned(posint)<g_list_length(GTK_BOX(box)->children)-1){
		gtk_box_reorder_child(GTK_BOX(box), panel,
				posint+1);
	} else std::puts("Nop!");
}

// callback used by multipass_dialog to traverse all children of the vbox
// holding the pass panels 
void for_each_callback(GtkWidget* wid, void* userdata)
{
	std::map<GtkWidget*, pass*>::iterator it=::panel_pass->find(wid);
	if(it!=::panel_pass->end()){
		std::vector<pass*>& algo=*static_cast< std::vector<pass*>* >(userdata);
		algo.push_back(it->second);
		//::panel_pass->erase(it);
	}
}

/// Handler for the Add Files button
void add_files_button_handler(GtkWidget* wid, gpointer data)
{ 
	GtkWidget* dialog = gtk_file_chooser_dialog_new ("Add Files",
			GTK_WINDOW(::gimp_dialog),
			GTK_FILE_CHOOSER_ACTION_OPEN,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
			NULL);
	
	gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(dialog), true);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
		GSList *filenames;

		filenames = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER (dialog));
		while(filenames){
			std::printf("Got %s\n", (char*)(filenames->data));
			GtkTreeIter iter;
			gtk_list_store_append(list_store, &iter);
			gtk_list_store_set(list_store, &iter,
					column_filename, filenames->data,
					column_icon, GTK_STOCK_FILE,
					-1);
			g_free(filenames->data);
			filenames=g_slist_next(filenames);
			
		}
		g_slist_free(filenames);
	}

	gtk_widget_destroy (dialog);
}

/// Handler for the Remove Files button
void remove_files_button_handler(GtkWidget* wid, gpointer)
{
	GtkTreeSelection* selection=gtk_tree_view_get_selection(GTK_TREE_VIEW(::tree_view));
	GtkTreeModel* model;

	GList* selected_rows=gtk_tree_selection_get_selected_rows(selection, &model);
	
	// Convert selected rows into a list of GtkTreeRowReference
	{
		GList* selected_rows2=selected_rows;
		while(selected_rows2){
			GtkTreePath* const path=static_cast<GtkTreePath*>(selected_rows2->data);
			GtkTreeRowReference* ref=gtk_tree_row_reference_new(GTK_TREE_MODEL(::list_store),
					path);
			selected_rows2->data=ref;
			gtk_tree_path_free(path);
			selected_rows2=g_list_next(selected_rows2);

		}
	}

	while(selected_rows){
		GtkTreeRowReference* const ref=static_cast<GtkTreeRowReference*>(selected_rows->data);
		GtkTreePath* const path=gtk_tree_row_reference_get_path(ref);

		GtkTreeIter iter;
		gtk_tree_model_get_iter(GTK_TREE_MODEL(::list_store), &iter, path);
		gtk_list_store_remove(::list_store, &iter);

		gtk_tree_path_free(path);
		gtk_tree_row_reference_free(ref);
		selected_rows=g_list_next(selected_rows);
	}
	g_list_free(selected_rows);

	std::printf("Remove files\n"); 
}

/// Handler for the Save Files button
void save_files_button_handler(GtkWidget* wid, gpointer)
{
	GtkWidget *dialog = gtk_file_chooser_dialog_new ("Save Benchmark",
			GTK_WINDOW(::gimp_dialog),
			GTK_FILE_CHOOSER_ACTION_SAVE,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
			NULL);
	gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);

	if (benchmark_filename->empty())
	{
		gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "untitled_benchmark.list");
	}
	else
		gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog), ::benchmark_filename->c_str());

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
	{
		char* cfilename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		std::string const filename=cfilename;
		g_free (cfilename);

		FILE* fp=fopen(filename.c_str(), "w");
		if(!fp){
			GtkWidget* dialog=gtk_message_dialog_new(GTK_WINDOW(::gimp_dialog),
					GTK_DIALOG_MODAL,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_OK,
					"Could not open '%s': %s",
					filename.c_str(),
					std::strerror(errno));
			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
		} else {
			*::benchmark_filename=filename;
			GtkTreeIter iter;
			GValue value;
			if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(::list_store),
						&iter)){
				do {
					std::fill((char*)&value, (char*)&value+sizeof(GValue),
							0);
					gtk_tree_model_get_value(GTK_TREE_MODEL(::list_store),
							&iter,
							column_filename,
							&value);
					std::fprintf(fp, "%s\n",
							(char*)g_value_peek_pointer(&value));
					g_value_unset(&value);
				} while(gtk_tree_model_iter_next(GTK_TREE_MODEL(::list_store),
							&iter));
			}
			fclose(fp);
		}

	}

	gtk_widget_destroy (dialog);
}

/// Handler for the Open Files button
void open_files_button_handler(GtkWidget* wid, gpointer data)
{ 
	GtkWidget* dialog = gtk_file_chooser_dialog_new ("Open Benchmark",
			GTK_WINDOW(::gimp_dialog),
			GTK_FILE_CHOOSER_ACTION_OPEN,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
			NULL);

	//GtkListStore* list_store=GTK_LIST_STORE(data);
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
	{
		char* cfilename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		std::string const filename=cfilename;
		g_free (cfilename);

		FILE* fp=std::fopen(filename.c_str(), "r");
		if(!fp){
			GtkWidget* dialog=gtk_message_dialog_new(GTK_WINDOW(::gimp_dialog),
					GTK_DIALOG_MODAL,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_OK,
					"Could not open '%s': %s",
					filename.c_str(),
					std::strerror(errno));
			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
		} else {
			gtk_list_store_clear(::list_store);
			char buffer[300];
			GtkTreeIter iter;
			while(! std::feof(fp)){
				if(!fgets(buffer, sizeof(buffer), fp)) break;
				std::size_t sz=std::strlen(buffer);
				if(buffer[sz-1]=='\n') buffer[sz-1]=0;
				if(buffer[sz-2]=='\r') buffer[sz-2]=0;

				gtk_list_store_append(::list_store,
						&iter);
				gtk_list_store_set(list_store, &iter,
						column_filename, buffer,
						column_icon, GTK_STOCK_FILE,
						-1);
			}
			std::fclose(fp);
			*::benchmark_filename=filename;
		}


	}

	gtk_widget_destroy (dialog);
}

///  Creates and returns the bottom benchmark editor
GtkWidget* create_benchmark_editor()
{
	GtkListStore* list_store=::list_store=gtk_list_store_new(nr_column,
			G_TYPE_STRING, G_TYPE_STRING);


	::tree_view=gtk_tree_view_new_with_model(GTK_TREE_MODEL(::list_store));
	GtkTreeSelection* selection=gtk_tree_view_get_selection(GTK_TREE_VIEW(::tree_view));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

	gtk_tree_view_append_column(GTK_TREE_VIEW(::tree_view),
			gtk_tree_view_column_new_with_attributes(" ",
				gtk_cell_renderer_pixbuf_new(),
				"stock-id", column_icon, NULL));
	gtk_tree_view_append_column(GTK_TREE_VIEW(::tree_view),
			gtk_tree_view_column_new_with_attributes("Files",
				gtk_cell_renderer_text_new(),
				"text", column_filename, NULL));
	
	GtkWidget* file_entry=::result_file_entry=gimp_file_entry_new("Save results to", "", false, false);
	GtkWidget* file_entry_label=gtk_label_new("Save results to: ");
	GtkWidget* file_entry_box=gtk_hbox_new(false, 0);
	gtk_box_pack_start(GTK_BOX(file_entry_box), file_entry_label, false, false, 0);
	gtk_box_pack_start(GTK_BOX(file_entry_box), file_entry, true, true, 0);
	gtk_widget_show(file_entry);
	gtk_widget_show(file_entry_label);

	GtkWidget* toolbar_box=gtk_hbox_new(false, 10);
	GtkWidget* save_button=gtk_button_new_from_stock(GTK_STOCK_SAVE);
	g_signal_connect(G_OBJECT(save_button), "clicked",
			G_CALLBACK(save_files_button_handler), list_store);
	GtkWidget* open_button=gtk_button_new_from_stock(GTK_STOCK_OPEN);
	g_signal_connect(G_OBJECT(open_button), "clicked",
			G_CALLBACK(open_files_button_handler), list_store);
	GtkWidget* add_button=gtk_button_new_from_stock(GTK_STOCK_ADD);
	g_signal_connect(G_OBJECT(add_button), "clicked",
			G_CALLBACK(add_files_button_handler), list_store);
	GtkWidget* remove_button=gtk_button_new_from_stock(GTK_STOCK_REMOVE);
	g_signal_connect(G_OBJECT(remove_button), "clicked",
			G_CALLBACK(remove_files_button_handler), list_store);
	GtkWidget* sep=gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(toolbar_box), add_button, false, false, 0);
	gtk_box_pack_start(GTK_BOX(toolbar_box), remove_button, false, false, 0);
	gtk_box_pack_start(GTK_BOX(toolbar_box), sep, false, false, 0);
	gtk_box_pack_start(GTK_BOX(toolbar_box), save_button, false, false, 0);
	gtk_box_pack_start(GTK_BOX(toolbar_box), open_button, false, false, 0);
	gtk_widget_show(save_button);
	gtk_widget_show(sep);
	gtk_widget_show(open_button);
	gtk_widget_show(add_button);
	gtk_widget_show(remove_button);

	GtkWidget* main_vbox=gtk_vbox_new(false, 5);
	gtk_box_pack_start(GTK_BOX(main_vbox), file_entry_box, false, false, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), toolbar_box, false, false, 0);
	gtk_widget_show(file_entry_box);

	GtkWidget* scrolled=gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled), ::tree_view);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled), GTK_POLICY_AUTOMATIC,
			GTK_POLICY_AUTOMATIC);
	gtk_widget_show(::tree_view);
	
	gtk_box_pack_start(GTK_BOX(main_vbox), scrolled, true, true, 0);
	gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 10);
	gtk_widget_show(toolbar_box);
	gtk_widget_show(scrolled);

	GtkWidget* main_frame=gtk_frame_new("Benchmark");
	gtk_container_add(GTK_CONTAINER(main_frame), main_vbox);
	gtk_widget_show(main_vbox);

	return  main_frame;
}

// Do-nothing progress indicator
struct nop_progress_indicator : public progress_indicator
{ virtual void operator()(double) {} };

} // anonymous namespace

bool multipass_batch_dialog(std::vector<pass*>& algo,
		std::vector<std::string>& benchmark,
		std::string& result_filename)
{
	std::string benchmark_filename;
	std::map<GtkWidget*, pass*> panel_pass;
	::panel_pass=&panel_pass;
	::benchmark_filename=&benchmark_filename;
	
	GtkWidget* dialog=::gimp_dialog=gimp_dialog_new("Multipass Segmenetation (Batch Mode)",
			"Multipass Segmentation",
			NULL, GtkDialogFlags(0),
			gimp_standard_help_func, "plug-in-multipass-batch",

			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OK,     GTK_RESPONSE_OK,

			NULL);

	GtkWidget* vbox=::pass_panel_vbox=gtk_vbox_new(false, 5);

	for(std::vector<pass*>::iterator it=algo.begin();
			it!=algo.end(); ++it)
	{
		pass* p=(*it)->clone();
		GtkWidget* pp=pass_panel(p);
		gtk_box_pack_start(GTK_BOX(::pass_panel_vbox), pp, false, false, 0);
		gtk_widget_show(pp);
	}

	GtkWidget* scrolled=gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_container_set_border_width(GTK_CONTAINER(scrolled), 10);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);

	gtk_widget_show(vbox);
	gtk_widget_show(scrolled);

	GtkWidget* vbox2=gtk_vbox_new(false,0);

	GtkWidget* add_button=gtk_button_new_from_stock(GTK_STOCK_ADD);
	g_signal_connect(G_OBJECT(add_button), "clicked", G_CALLBACK(add_button_handler),
			vbox);
	gtk_widget_show(add_button);

	gtk_box_pack_start(GTK_BOX(vbox2), vbox, true, true, 0);
	gtk_box_pack_end(GTK_BOX(vbox2), add_button, false, false, 10);
	gtk_widget_show(vbox2);

	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled), vbox2);
	
	GtkWidget* frame=gtk_frame_new("Algorithm");
	gtk_container_add(GTK_CONTAINER(frame), scrolled);
	gtk_widget_show(frame);

	GtkWidget* benchmark_editor=create_benchmark_editor();

	GtkWidget* hpane=gtk_vpaned_new();
	gtk_paned_pack1(GTK_PANED(hpane), frame, true, true);
	gtk_paned_pack2(GTK_PANED(hpane), benchmark_editor, false, false);
	gtk_widget_show(benchmark_editor);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
			hpane, true, true, 10);
	gtk_widget_show(hpane);
	
	gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 10);

	gtk_window_set_default_size(GTK_WINDOW(dialog), 600, 500);

	while(gimp_dialog_run(GIMP_DIALOG(dialog))==GTK_RESPONSE_OK){
		char* cfilename=gimp_file_entry_get_filename(GIMP_FILE_ENTRY(::result_file_entry));
		result_filename=cfilename;
		g_free(cfilename);

		GtkTreeIter iter;
		if(result_filename.empty()){
			GtkWidget* dialog=gtk_message_dialog_new(GTK_WINDOW(::gimp_dialog),
					GTK_DIALOG_MODAL,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_OK,
					"Please provide a correct output file for"
					" results");
			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
		} else if (! gtk_tree_model_get_iter_first(GTK_TREE_MODEL(::list_store), &iter)) {
			GtkWidget* dialog=gtk_message_dialog_new(GTK_WINDOW(::gimp_dialog),
					GTK_DIALOG_MODAL,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_OK,
					"Please provide a benchmark (list of files)");
			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
		} else {
			// Delete old ones
			for(std::vector<pass*>::iterator it=algo.begin(); it!=algo.end();
					++it)
				delete *it;
			algo.clear();

			// Insert new ones in algo inorder
			gtk_container_foreach(GTK_CONTAINER(vbox), for_each_callback,
					&algo);

			benchmark.clear();
			GValue value;
			do {
				std::fill((char*)&value, (char*)&value+sizeof(GValue),
						0);
				gtk_tree_model_get_value(GTK_TREE_MODEL(::list_store),
						&iter,
						column_filename,
						&value);
				benchmark.push_back((char*)g_value_peek_pointer(&value));
				g_value_unset(&value);
			}while (gtk_tree_model_iter_next(GTK_TREE_MODEL(::list_store), &iter));

			gtk_widget_destroy(dialog);
			return true;
		}
	}
	// delete new ones
	for(std::map<GtkWidget*, pass*>::iterator it=::panel_pass->begin();
			it!=::panel_pass->end(); ++it){
		delete it->second;
	}
	gtk_widget_destroy(dialog);
	return false;
}
