/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <iterator>
#include <map>
#include <cstdio>
#include <iostream>
#include <sstream>
#include <cstring>

#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "multipass_dialog.hpp"
#include "pixel_class.hpp"
#include "colors.hpp"
#include "ui_build.hpp"
#include "pixel_region.hpp"
#include "pass_repertoire.hpp"
#include "rgb_to_gray.hpp"
#include "compute_histogram.hpp"
#include "progress_indicator.hpp"
#include "fitness_function_repertoire.hpp"
#include "pixel_classifier.hpp"
#include "pixel_classifier_repertoire.hpp"
#include "jaccard_distance.hpp"
#include "load_manual_segmentation.hpp"

namespace
{
char const * const jaccard_label="Jaccard (Supervised)";

int jaccard_index;

void remove_button_handler(GtkWidget* widget, gpointer data);
void up_button_handler(GtkWidget* button, gpointer data);
void down_button_handler(GtkWidget* button, gpointer data);

// Matches an entry panel (as returned by ui_build)
// to the pass uppon which it is acting
std::map<GtkWidget*, pass*>* panel_pass;

// The preview widget
GtkWidget* preview;

// The drawable we're (going to) work(ing) on
GimpDrawable* drawable;

// The manual segmentation of ::drawable when available
GimpDrawable* ground_truth_drawable;

// The vbox containing a panel per pass
GtkWidget* pass_panel_vbox;

// Wether to update the preview
gboolean update_preview=true;

// For the last segmentation preview a fitness value for each fitness function
std::vector<double> *fitnesses;

// The progress bar showing the current value of the
// selected fitness function
GtkWidget* fitness_indicator;

// The currently shown fitness index
int active_fitness;

/// The currently chosen pixel_classifier
pixel_classifier* current_pixel_classifier;

// updates ::fitness_indicator to show ::active_fitness index value
void update_fitness_indicator()
{
	double const value=(*fitnesses)[active_fitness]<1?(*fitnesses)[active_fitness]:1;
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(fitness_indicator),value);
	std::ostringstream oss; oss<<(*fitnesses)[active_fitness];
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(fitness_indicator),
			oss.str().c_str());
}

// handler for the "changed" signal of the pixel_classifier combo
void pixel_classifier_combo_changed_handler(GtkWidget* combo, gpointer)
{
	int active=gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
	if(active<0) return;

	pixel_classifier_repertoire& rep=*pixel_classifier_repertoire::instance();
	for(pixel_classifier_repertoire::iterator it=rep.begin(); it!=rep.end(); ++it){
		if(active--==0) ::current_pixel_classifier=*it;	
	}
	gimp_preview_invalidate(GIMP_PREVIEW(::preview));
}

// Creates and returns a combo with an item per possible pixel_classifier
GtkWidget* pixel_classifier_combo()
{
	GtkWidget* combo=gtk_combo_box_new_text();

	pixel_classifier_repertoire& rep=*pixel_classifier_repertoire::instance();
	for(pixel_classifier_repertoire::iterator it=rep.begin();
			it!=rep.end(); ++it){
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo),
				(*it)->label());
	}
	g_signal_connect(G_OBJECT(combo), "changed",
			G_CALLBACK(pixel_classifier_combo_changed_handler),
			0);

	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
	return combo;
}

// handler for the "changed" signal of the fitness combo
void fitness_combo_changed_handler(GtkWidget* combo, gpointer)
{
	int active=gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
	if(active<0) return;
	::active_fitness=active;
	update_fitness_indicator();
}

// Creates and returns a combo with an item per possible fitness
GtkWidget* fitness_combo()
{
	GtkWidget* combo=gtk_combo_box_new_text();

	fitness_function_repertoire& rep=*fitness_function_repertoire::instance();
	for(fitness_function_repertoire::iterator it=rep.begin();
			it!=rep.end(); ++it){
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo),
				(*it)->label());
	}
	if(::ground_truth_drawable){
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo), jaccard_label);
		::jaccard_index=rep.size()+1;
	}
	
	g_signal_connect(G_OBJECT(combo), "changed",
			G_CALLBACK(fitness_combo_changed_handler),
			0);

	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
	return combo;
}

// handler for "value-changed" of slides and scales
void value_changed_handler(GtkWidget* value, gpointer)
{
	gimp_preview_invalidate(GIMP_PREVIEW(preview));
}

// handler for the combo box for selecting the pass type
void pass_combo_box_handler(GtkWidget* combo, gpointer data)
{
	GtkWidget* parent=GTK_WIDGET(data);
	GtkWidget* pass_ui=GTK_BIN(parent)->child;

	int active=gtk_combo_box_get_active(GTK_COMBO_BOX(combo));

	if(active<0)active=0;

	pass_repertoire::iterator it=pass_repertoire::instance()->begin();
	std::advance(it, active);

	pass* new_pass=(*it)->clone();
	GtkWidget* new_pass_ui=ui_build(new_pass, value_changed_handler);
	
	std::map<GtkWidget*, pass*>::iterator jt=panel_pass->find(
			gtk_widget_get_parent(gtk_widget_get_parent(pass_ui)));
	if(jt!=panel_pass->end()){
		delete jt->second;
		jt->second=new_pass;
	}

	gtk_widget_destroy(pass_ui);

	gtk_container_add(GTK_CONTAINER(parent), new_pass_ui);
	gtk_widget_show(new_pass_ui);
	gimp_preview_invalidate(GIMP_PREVIEW(::preview));
}

// Creates and returns a combo_box that contains all available pass' labels
// initial selection is set to passed string if possible
GtkWidget* label_chooser(char const* init, GtkWidget* pass_param_parent)
{
	pass_repertoire& rep=*pass_repertoire::instance();
	GtkWidget* combo=gtk_combo_box_new_text();

	gint index=-1;
	int i=0;
	for(pass_repertoire::iterator it=rep.begin();
			it!=rep.end(); ++it, ++i){
		if(std::strcmp((*it)->label(), init)==0)
			index=i;
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo),
				(*it)->label());
	}
	if(index>=0)
		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), index);
	else 
		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
	g_signal_connect(G_OBJECT(combo), "changed",
			G_CALLBACK(pass_combo_box_handler),
			pass_param_parent);
	return combo;
}

// Creates and returns a panel for a pass
GtkWidget* pass_panel(pass* p)
{
	GtkWidget* frame=gtk_frame_new("");
	
	GtkWidget* pass_param=ui_build(p, value_changed_handler);
	panel_pass->insert(std::make_pair(frame, p));

	GtkWidget* exp=gtk_expander_new("Parameters");
	
	GtkWidget* title_bar=gtk_hbox_new(false, 20);
	// Fill the title_bar
	{
		GtkWidget* label_ch=label_chooser(p->label(), exp);
		gtk_box_pack_start(GTK_BOX(title_bar),
				label_ch, false, false, 0);
		GtkWidget* button=gtk_button_new_from_stock(GTK_STOCK_REMOVE);
		GtkWidget* up_button=gtk_button_new_from_stock(GTK_STOCK_GO_UP);
		GtkWidget* down_button=gtk_button_new_from_stock(GTK_STOCK_GO_DOWN);
		g_signal_connect(G_OBJECT(button), "clicked", 
				G_CALLBACK(remove_button_handler),
				frame);
		g_signal_connect(G_OBJECT(up_button),
				"clicked",
				G_CALLBACK(up_button_handler),
				frame);
		g_signal_connect(G_OBJECT(down_button),
				"clicked",
				G_CALLBACK(down_button_handler),
				frame);
		gtk_box_pack_end(GTK_BOX(title_bar),
				button, false, false, 0);
		gtk_box_pack_end(GTK_BOX(title_bar),
				down_button, false, false, 0);
		gtk_box_pack_end(GTK_BOX(title_bar),
				up_button, false, false, 0);
		gtk_widget_show(label_ch);
		gtk_widget_show(button);
		gtk_widget_show(up_button);
		gtk_widget_show(down_button);
	}

	gtk_container_set_border_width(GTK_CONTAINER(title_bar), 10);

	gtk_frame_set_label_widget(GTK_FRAME(frame), title_bar);
	gtk_widget_show(title_bar);

	gtk_container_add(GTK_CONTAINER(exp), pass_param);
	gtk_widget_show(pass_param);

	gtk_container_set_border_width(GTK_CONTAINER(exp), 5);
	gtk_container_add(GTK_CONTAINER(frame), exp);
	gtk_widget_show(exp);
	gtk_widget_show(frame);
	
	return frame;
}

// Handler of the "+ Add" button
void add_button_handler(GtkWidget* widget, gpointer data)
{
	GtkWidget* vbox=GTK_WIDGET(data);
	
	pass* p=*pass_repertoire::instance()->begin();
	GtkWidget* panel=pass_panel(p->clone());
	gtk_box_pack_start(GTK_BOX(vbox), panel, 0, 0, false);
	gtk_widget_show(panel);

	gimp_preview_invalidate(GIMP_PREVIEW(::preview));
}

// Handler of the "- Remove button"
void remove_button_handler(GtkWidget* widget, gpointer data)
{
	GtkWidget* panel=GTK_WIDGET(data);
	GtkWidget* parent=gtk_widget_get_parent(panel);
	
	if(::panel_pass->size()==1) return;

	std::map<GtkWidget*, pass*>::iterator it=::panel_pass->find(panel);
	if(it!=panel_pass->end()){
		delete it->second;
		panel_pass->erase(it);
	}
	gtk_container_remove(GTK_CONTAINER(parent),
			panel);
	gimp_preview_invalidate(GIMP_PREVIEW(::preview));
}

// Handler of the Up button
void up_button_handler(GtkWidget* button, gpointer data)
{
	GtkWidget* panel=GTK_WIDGET(data);
	GtkWidget* box=gtk_widget_get_parent(panel);

	GValue pos;
	std::fill((char*)&pos, (char*)(&pos+1), 0);

	g_value_init(&pos, G_TYPE_INT);

	gtk_container_child_get_property(GTK_CONTAINER(box),
			panel, "position", &pos);

	int posint=g_value_get_int(&pos);

	if(posint>0){
		gtk_box_reorder_child(GTK_BOX(box), panel,
				posint-1);
		gimp_preview_invalidate(GIMP_PREVIEW(::preview));
	}
}

// Handler of the down button
void down_button_handler(GtkWidget* button, gpointer data)
{
	GtkWidget* panel=GTK_WIDGET(data);
	GtkWidget* box=gtk_widget_get_parent(panel);

	GValue pos;
	std::fill((char*)&pos, (char*)(&pos+1), 0);

	g_value_init(&pos, G_TYPE_INT);

	gtk_container_child_get_property(GTK_CONTAINER(box),
			panel, "position", &pos);

	int posint=g_value_get_int(&pos);
	
	if(unsigned(posint)<g_list_length(GTK_BOX(box)->children)-1){
		gtk_box_reorder_child(GTK_BOX(box), panel,
				posint+1);
		gimp_preview_invalidate(GIMP_PREVIEW(::preview));
	} else std::puts("Nop!");
}

// callback used by multipass_dialog to traverse all children of the vbox
// holding the pass panels 
void for_each_callback(GtkWidget* wid, void* userdata)
{
	std::map<GtkWidget*, pass*>::iterator it=::panel_pass->find(wid);
	if(it!=::panel_pass->end()){
		std::vector<pass*>& algo=*static_cast< std::vector<pass*>* >(userdata);
		algo.push_back(it->second);
		//::panel_pass->erase(it);
	}
}

// Do-nothing progress indicator
struct nop_progress_indicator : public progress_indicator
{ virtual void operator()(double) {} };
#if 0
namespace {
/// RGB trio
struct rgb { guchar r, g, b; };

// Fills a vector with random colors
void fill_colors(std::vector<rgb>& colors)
{
	for(std::vector<rgb>::iterator it=colors.begin();
			it!=colors.end(); ++it){
		it->r=std::rand()%256;
		it->g=std::rand()%256;
		it->b=std::rand()%256;
	}
}

std::vector<rgb> class_colors(256);

} // anonymous namespace
#endif

// Handler of the invalidate event of the GimpDrawable
void invalidated_handler(GtkWidget* wid, gpointer data)
{
	bool const debug=false;

	if(debug) std::printf("invalidate!\n");
	
	gint x, y, w, h;
	gimp_preview_get_position(GIMP_PREVIEW(::preview), &x, &y);
	gimp_preview_get_size(GIMP_PREVIEW(::preview), &w, &h);
	
	gimp_tile_cache_ntiles (2 * (w / 
                                     gimp_tile_width () + 1));	

	histogram hist;
	compute_histogram(drawable, &hist, x, y, w, h);

	std::vector<pass*> algo;
	// Insert in algo inorder
	gtk_container_foreach(GTK_CONTAINER(::pass_panel_vbox), for_each_callback,
			&algo);

	if(debug)for(std::vector<pass*>::iterator it=algo.begin(); it!=algo.end(); ++it)
		(*it)->print();

	centroid_set centroids;
	nop_progress_indicator ind;
	for(std::vector<pass*>::iterator it=algo.begin(); it!=algo.end(); ++it)
		(*it)->operator() (hist, centroids, &ind);

	// Compute fitnesses
	classification c12n;
	{
		classify(hist, centroids, &c12n);

		fitness_function_repertoire& rep=*fitness_function_repertoire::instance();
		fitness_function_repertoire::iterator rt=rep.begin();
		for(std::vector<double>::iterator it=::fitnesses->begin();
				it!=::fitnesses->end()-1; ++it)
		{
			*it=(*rt)->operator()(hist, centroids, c12n);
			++rt;
		}

		::update_fitness_indicator();	
	}		

	// Compute supervised evaluation
	if(::ground_truth_drawable) do {
		::fitnesses->back()=0.0; // fall back value

		histogram supervised_histogram;
		compute_histogram(::ground_truth_drawable, &supervised_histogram,
				x, y, w, h);
	
		for(histogram::iterator it=supervised_histogram.begin(); it!=
				supervised_histogram.end(); ++it){
			if(*it) std::cout<<it-supervised_histogram.begin()<<" ";
		}
		std::cout<<"\n";
		
		std::vector<std::pair<int, double> > out;
		jaccard_distance(::drawable, c12n,  centroids, ::ground_truth_drawable,
			supervised_histogram, x, y, w, h, &out); 
	
		// Computes the mean of all jaccard indexes
		double mean=0.0;
		for(std::vector<std::pair<int, double> >::iterator it=out.begin();
				it!=out.end(); ++it){
				mean+=it->second;
			std::cout<<it->first<<" "<<it->second<<"\n";
		}
		std::cout<<"-------------------------\n";
		mean/=out.size();
		::fitnesses->back()=mean;
	} while(0); // end of compute supervised evaluation

	GimpPixelRgn rgn_out[1];
	gimp_pixel_rgn_init (rgn_out,
			drawable,
			x, y,
			w, h, 
			false, true);
	
	GimpPixelRgn rgn_in[1];
	gimp_pixel_rgn_init (rgn_in,
			drawable,
			x, y,
			w, h, false, false);

	std::vector<guchar> row(w*::drawable->bpp);
	for(int j=y; j<h+y; ++j){
		gimp_pixel_rgn_get_row(rgn_in,
				&row[0], x, j, w);
		for(std::vector<guchar>::iterator it=row.begin(); it!=row.end();
				it+=::drawable->bpp){
			int const pclass=::current_pixel_classifier->operator() (hist, centroids,
					c12n, ::drawable, rgn_in, *it, x+(it-row.begin())/::drawable->bpp, j,
					1.0
					);
			if(::drawable->bpp==1){
				*it=guchar(centroids[pclass]);
			} else {
				rgb const color=colors[pclass%256];
				it[0]=color.r;it[1]=color.g;it[2]=color.b;
				if(::drawable->bpp==4) it[3]=255;
			}
		}
		gimp_pixel_rgn_set_row(rgn_out,
				&row[0],
				x, j, w);
	}

	gimp_drawable_preview_draw_region(GIMP_DRAWABLE_PREVIEW(preview),
			rgn_out);
}

} // anonymous namespace

bool multipass_dialog(GimpDrawable*& drawable, std::vector<pass*>& algo,
		pixel_classifier** pixel_cls)
{
	std::map<GtkWidget*, pass*> panel_pass;
	std::vector<double> fitnesses(fitness_function_repertoire::instance()->size()+1);

	::fitnesses=&fitnesses;

	::drawable=drawable;
	::panel_pass=&panel_pass;
	
	// Look for corresponding manual segmentation and fill
	// ::ground_truth_drawable correspondingly 
	::ground_truth_drawable=load_manual_segmentation(gimp_drawable_get_image(drawable->drawable_id));
	
	GtkWidget* dialog=gimp_dialog_new("Multipass Segmenetation",
			"Multipass Segmentation",
			NULL, GtkDialogFlags(0),
			gimp_standard_help_func, "plug-in-multipass",

			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OK,     GTK_RESPONSE_OK,

			NULL);

	/// Make sure image is RGB
	bool converted_to_rgb=false;
	if(drawable->bpp<3){
		gint32 image_id=gimp_drawable_get_image(drawable->drawable_id);
		gimp_drawable_detach(drawable);
		converted_to_rgb=gimp_image_convert_rgb(image_id);
		::drawable=drawable=gimp_drawable_get(gimp_image_get_active_drawable(image_id));
	}

	/// Test code (intended to be revamped or removed
	if(false) {
		gint32 preview_image_id=gimp_image_new(drawable->width, drawable->height, GIMP_RGB);
		gint32 preview_drawable_id=gimp_layer_new(preview_image_id, "Preview", drawable->width,
				drawable->height, GIMP_RGB_IMAGE, 1.0, GIMP_NORMAL_MODE);
		gimp_image_add_layer(preview_image_id, preview_drawable_id, 0);
		// Copies drawable content into preview_drawable_id
		{
			GimpDrawable* preview_drawable=gimp_drawable_get(preview_drawable_id);
			pixel_region in(drawable, 0, 0, drawable->width, drawable->height,
					false, false);
			pixel_region out(preview_drawable, 0, 0, drawable->width, drawable->height,
					true, true);
			pixel_region::iterator out_it=out.begin();
			for(pixel_region::iterator in_it=in.begin(); in_it!=in.end(); ++in_it){
				pixel_region::chunk const& in_chunk=*in_it;
				guchar* out_ch=out_it->begin;
				for(guchar* ch=in_chunk.begin; ch!=in_chunk.end; ch+=drawable->bpp){
					if(drawable->bpp==1){
						out_ch[0]=out_ch[1]=out_ch[2]=*ch;
					} else {
						out_ch[0]=ch[0];
						out_ch[1]=ch[1];
						out_ch[2]=ch[2];
					}
					out_ch+=preview_drawable->bpp;
				}

			}
			gimp_drawable_merge_shadow(preview_drawable_id, false);

			gimp_drawable_detach(preview_drawable);
			gimp_display_new(preview_image_id);
		}
	}// end of test code

	::preview=gimp_drawable_preview_new(drawable, &::update_preview);
	g_signal_connect(G_OBJECT(preview), "invalidated",
			G_CALLBACK(invalidated_handler),
			0);

	GtkWidget* preview_frame=gtk_frame_new("Preview");
	GtkWidget* preview_box=gtk_vbox_new(false, 10);
	::fitness_indicator=gtk_progress_bar_new();
	GtkWidget* fitness_chooser=fitness_combo();
	GtkWidget* performance_box=gtk_vbox_new(false, 0);
	gtk_container_set_border_width(GTK_CONTAINER(performance_box), 10);
	GtkWidget* performance_frame=gtk_frame_new("Performance");

	gtk_box_pack_start(GTK_BOX(performance_box), fitness_chooser, false, false, 0);
	gtk_box_pack_start(GTK_BOX(performance_box), ::fitness_indicator, false, false, 0);
	gtk_widget_show(fitness_chooser);
	gtk_widget_show(::fitness_indicator);

	gtk_container_add(GTK_CONTAINER(performance_frame), performance_box);
	gtk_widget_show(performance_box);

	gtk_box_pack_start(GTK_BOX(preview_box), preview, false, false, 0);
	gtk_box_pack_start(GTK_BOX(preview_box), performance_frame, false, false, 0);
	gtk_widget_show(preview);
	gtk_widget_show(performance_frame);

	gtk_container_add(GTK_CONTAINER(preview_frame), preview_box);
	gtk_container_set_border_width(GTK_CONTAINER(preview_box), 10);
	gtk_widget_show(preview_box);

	GtkWidget* vbox=::pass_panel_vbox=gtk_vbox_new(false, 5);

	GtkWidget* hbox=gtk_hbox_new(false, 10);

	for(std::vector<pass*>::iterator it=algo.begin();
			it!=algo.end(); ++it)
	{
		pass* p=(*it)->clone();
		GtkWidget* pp=pass_panel(p);
		gtk_box_pack_start(GTK_BOX(vbox), pp, false, false, 0);
		gtk_widget_show(pp);
	}

	GtkWidget* scrolled=gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_container_set_border_width(GTK_CONTAINER(scrolled), 10);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);

	gtk_widget_show(vbox);
	gtk_widget_show(scrolled);

	GtkWidget* vbox2=gtk_vbox_new(false,0);

	GtkWidget* add_button=gtk_button_new_from_stock(GTK_STOCK_ADD);
	g_signal_connect(G_OBJECT(add_button), "clicked", G_CALLBACK(add_button_handler),
			vbox);
	gtk_widget_show(add_button);

	gtk_box_pack_start(GTK_BOX(vbox2), vbox, true, true, 0);
	gtk_box_pack_end(GTK_BOX(vbox2), add_button, false, false, 10);
	gtk_widget_show(vbox2);

	gtk_box_pack_start(GTK_BOX(hbox), preview_frame, false, false, 0);
	gtk_widget_show(preview_frame);
	
	GtkWidget* algo_frame=gtk_frame_new("Algorithm");
	gtk_container_add(GTK_CONTAINER(algo_frame), scrolled);
	gtk_widget_show(scrolled);

	GtkWidget* algo_pixel_cls_vbox=gtk_vbox_new(false, 10);
	
	GtkWidget* pixel_cls_frame=gtk_frame_new("Pixel Classifier");
	GtkWidget* pixel_cls_vbox=gtk_vbox_new(false, 0);
	GtkWidget* pixel_cls_combo=pixel_classifier_combo();
	gtk_box_pack_start(GTK_BOX(pixel_cls_vbox), pixel_cls_combo, false, false, 0);
	gtk_widget_show(pixel_cls_combo);
	gtk_container_add(GTK_CONTAINER(pixel_cls_frame), pixel_cls_vbox);
	gtk_widget_show(pixel_cls_vbox);
	gtk_container_set_border_width(GTK_CONTAINER(pixel_cls_vbox), 10);

	gtk_box_pack_start(GTK_BOX(algo_pixel_cls_vbox), algo_frame, true, true, 0);
	gtk_box_pack_end(GTK_BOX(algo_pixel_cls_vbox), pixel_cls_frame, false, false, 0);
	gtk_widget_show(algo_frame);
	gtk_widget_show(pixel_cls_frame);

	gtk_box_pack_end(GTK_BOX(hbox), algo_pixel_cls_vbox, true, true, 0);
	gtk_widget_show(algo_pixel_cls_vbox);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
			hbox, true, true, 0);
	
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled), vbox2);
	
	gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 10);

	gtk_window_set_default_size(GTK_WINDOW(dialog), 800, 400);
	gtk_widget_show(hbox);

	bool const ret=(gimp_dialog_run(GIMP_DIALOG(dialog))==GTK_RESPONSE_OK);
	if(ret){
		// Delete old ones
		for(std::vector<pass*>::iterator it=algo.begin(); it!=algo.end();
				++it)
			delete *it;
		algo.clear();

		// Insert new ones in algo inorder
		gtk_container_foreach(GTK_CONTAINER(vbox), for_each_callback,
				&algo);
		*pixel_cls=::current_pixel_classifier;
	
	} else {
		// delete new ones
		for(std::map<GtkWidget*, pass*>::iterator it=::panel_pass->begin();
				it!=::panel_pass->end(); ++it){
			delete it->second;
		}
	}
	gtk_widget_destroy(dialog);
	
	drawable=::drawable;
	return ret;
}
