/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef PARAM_HPP
#define PARAM_HPP

#include <new>

/// A parameter as passed to an algorithm/function
///
/// This class' main purpose is to allow dynamic parameter binding (that is to
/// be able to pass parameters whose type is not known at compile time).
/// To this end, the class provides storage and access to many possible C++
/// types (an instance may hold a double, an int..etc) and a way to know about
/// the actual type held. (this is a much simpler/faster/basic version of
/// boost::any) 
struct param
{
	/// The actual type of the parameter
	struct type_info
	{ 
		enum type {
			int_type, double_type
		};
	};

	/// Create an integer parameter with the given value
	param(int i): m_type(type_info::int_type),
		m_int_value(i){}
	
	/// Create a double parameter with the given value
	param(double i): m_type(type_info::double_type),
		m_double_value(i){}

	param(const param& p)
	{
		this->m_type=p.m_type;
		switch(this->m_type){
			case type_info::int_type:
				this->m_int_value=p.m_int_value; break;
			case type_info::double_type: 
				this->m_double_value=p.m_double_value; break;
		}
	}

	param& operator=(param const& p){ return *new (this) param(p); }

	bool operator==(param const& p) const
	{
		if(m_type!=p.m_type)return false;
		switch(m_type){
			case type_info::int_type:
				if(int_value()!=p.int_value()) 
					return false;
				return true;
			case type_info::double_type:
				if(double_value()!=p.double_value()) 
					return false;
				return true;
		}
		return true;
	}
			
	/// Access to the int value
	int& int_value() { return m_int_value; }
	int const& int_value() const { return m_int_value; }
	
	/// Access to the double value
	double& double_value() { return m_double_value; }
	double const& double_value() const { return m_double_value; }

	/// Access to the type of the parameter
	type_info::type type() const { return m_type; }
private:
	type_info::type m_type;
	union {
		int m_int_value;
		double m_double_value;
	};
};

/// Information about a parameter to be passed to an algorithm/function
///
/// A param_info describes a parameter/data you_need/is_expected to be passed to
/// an algorithm. The first information a param_info holds is the type (C++
/// speaking) of the data. There is an enumeration value of
/// param::type_info::type for each possible/supported data type. Use
/// param_info::type() to get it. For example, if param_info::type() yields
/// param::type_info::int_type it means the data expected is an integer.
/// 
/// Next, for each possible data type, a param_info holds additional
/// information. For instance, a param_info whose param_info::type() yields
/// param::type_info::double_type, also holds the _range_ in which that double is
/// expected (see param_info::double_start()/double_end())(the same goes for int
/// ranges.
///
/// Additionally a param_info holds some information (like a label, and a
/// description) that are always present regardless of the type of the data.
struct param_info 
{

	/// Create a parameter info containing a range of integers
	param_info(int start, int end, char const* lab,
			char const* desc=0):
		m_label(lab),
		m_description(desc),
		m_type(param::type_info::int_type)
	{
		int_type_info.start=start;
		int_type_info.end=end;
	}

	/// Create a parameter info containing a range of doubles
	param_info(double start, double end, char const* lab,
			char const* desc=0):
		m_label(lab),
		m_description(desc),
		m_type(param::type_info::double_type)
	{
		double_type_info.start=start;
		double_type_info.end=end;
	}
	
	/// Access to the start of a double range
	double& double_start() { return double_type_info.start; }
	double const& double_start() const { return double_type_info.start; }
	
	/// Access to the end of a double range
	double& double_end() { return double_type_info.end; }
	double const& double_end() const { return double_type_info.end; }

	/// Access to the start of a int range
	int& int_start() { return int_type_info.start; }
	int const& int_start() const { return int_type_info.start; }
	
	/// Access to the end of a int range
	int& int_end() { return int_type_info.end; }
	int const& int_end() const { return int_type_info.end; }

	/// Access to the label
	char const*& label() { return m_label; }
	char const* const& label() const { return m_label; }

	/// Access to the description
	char const*& description() { return m_description; }
	char const* const& description() const { return m_description; }

	/// Access to the actual type of the parameter (information)
	param::type_info::type type() const { return m_type; }

	private:
	/// Type-specific data about the parameter
	union {
		/// Range of possible values for an int_type parameter
		struct {
			int start,end;
		} int_type_info;

		/// Range of possible values for a double_type parameter
		struct {
			double start,end;
		} double_type_info;
	};

	char const* m_label;
	char const* m_description;
	param::type_info::type m_type;

};

#endif // PARAM_HPP
