/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef PASS_HPP
#define PASS_HPP

#include <cstdlib>
#include <vector>
#include <algorithm>

#include "param.hpp"
#include "histogram.hpp"
#include "centroid_set.hpp"
#include "progress_indicator.hpp"

/// <Abstract> A step in an algorithm. A pass can be used stand alone and
/// considered an alroithm by itself, but this abstract base class serves
/// to treat a complex algorithm as a collection of passes each doing something
/// different. 
///
/// Child classes are supposed to provide the following (through proper
/// overloading)
/// - A description of the type of the parameters that the pass requests;
/// - The actual implementation of the (sub)-algorithm
///
/// Class pass itself will provide child classes with storage for their
/// parameters and easy access to them (through and index)
///
/// This implements the prototype pattern (through a pure virtual clone()
/// member)
class pass
{
	public:
	/// \param pt The address of an array of param_info which describes the
	/// parameters needed by the pass
	/// \param pt_size number of elements in pt
	/// The passed array is not copied internally and is not expected to be
	/// altered after the call. 
	///
	/// The initial values of the parameters (as return by parameter_value)
	/// will be the first/lowest valid value for each parameter.
	pass(char const* label, param_info const* pt, std::size_t pt_size);

	/// Perform operation (to be overloaded).
	///
	/// Passed parameters are the histogram of the image and initial
	/// position of the centroids, expected to return the new position of
	/// the centroids right in the passed reference.
	///
	/// The progress_indicator should be used by child classes to report the
	/// advancement of the operations.
	virtual
	void operator() (histogram const&, 
			centroid_set& centroids,
			progress_indicator* ind) const=0;

	/// Access to the current value of a parameter (by index)
	param& parameter_value(std::size_t index) 
	{ return m_parameters_value[index];}

	/// Access to the current value of a parameter (by index)
	param const& parameter_value(std::size_t index) const 
	{ return m_parameters_value[index];}
	
	/// Get information about a parameter (by index)
	param_info const& parameter_info(std::size_t index) const
	{ return m_parameters_info[index]; }
	
	/// Number of parameters accepted/expected
	std::size_t nr_parameters() const { return m_nr_parameters; }

	/// Access to the label
	char const* label() const
	{ return m_label; }

	/// For debugging purposes
	virtual void print() const{};

	virtual ~pass(){}

	/// Child class should override this to return newed copies of
	/// themselves
	virtual pass* clone() const=0;

	protected:
	/// Copies parameter data (should be useful only to child classes when
	/// implementing clone()
	//pass(pass const& p);

	private:
	pass& operator=(pass const&);

	param_info const* m_parameters_info;
	std::size_t m_nr_parameters;
	std::vector<param> m_parameters_value;
	char const* m_label;
};


#endif // PASS_HPP
