/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <algorithm>
#include <cassert>

#include "pixel_region.hpp"

pixel_region::pixel_region(GimpDrawable* dr,
		gint x, gint y, gint w, gint h,
		bool shadow, bool dirty):
	m_drawable(dr), m_x(x), m_y(y), m_width(w),
	m_height(h), m_shadow(shadow), m_dirty(dirty)
{
}

pixel_region::iterator::iterator(pixel_region const& pr):
	m_region(&pr), m_end_flag(false)
{
	guint const tile_col=pr.x()/gimp_tile_width();
	guint const tile_row=pr.y()/gimp_tile_height();
	move_to_tile(tile_row*pr.drawable()->ntile_cols+tile_col);
}

pixel_region::iterator::iterator(pixel_region const& pr,
		pixel_region::iterator::end_tag):
	m_region(&pr), m_end_flag(true)
{
	// move to last tile
	guint const tile_col=(pr.x()+pr.width()-1)/gimp_tile_width();
	guint const tile_row=(pr.y()+pr.height()-1)/gimp_tile_height();
	move_to_tile(tile_row*pr.drawable()->ntile_cols+tile_col);
}

pixel_region::iterator& pixel_region::iterator::operator++()
{
	if(m_end_flag) return *this;

	if(m_nr_chunks>0){
		// move to next chunk within current tile
		m_chunk.begin+=m_chunk_skip;
		m_chunk.end+=m_chunk_skip;
		m_nr_chunks--;
	} else {
		// try to move to next tile
		guint const next_tile_num=next_tile_number();
		if(next_tile_num!=m_tile->tile_num){
			gimp_tile_unref(m_tile, m_region->dirty());
			move_to_tile(next_tile_num);
		} else {
			// no more tiles, dead end
			m_end_flag=true;
		}
	}

	return *this;
}

bool pixel_region::iterator::operator==(pixel_region::iterator const& lhs) const
{
	if(this->m_region != lhs.m_region) return false;
	if (this->m_end_flag==true && lhs.m_end_flag==true) return true;
	if (this->m_end_flag != lhs.m_end_flag) return false;
	return m_tile->data==lhs.m_tile->data && m_nr_chunks==lhs.m_nr_chunks;
}

pixel_region::iterator::iterator(pixel_region::iterator const& lhs)
{
	m_region=lhs.m_region;
	
	m_tile=lhs.m_tile;
	gimp_tile_ref(m_tile);

	m_chunk=lhs.m_chunk;
	m_chunk_skip=lhs.m_chunk_skip;
	m_nr_chunks=lhs.m_nr_chunks;
	m_end_flag=lhs.m_end_flag;
}

guint pixel_region::iterator::next_tile_number() const
{
	guint const tilenum=m_tile->tile_num;
	guint const tile_col=tilenum%m_region->drawable()->ntile_cols;
	guint const tile_row=tilenum/m_region->drawable()->ntile_cols;

	// Position of tile in the drawable
	gint const m_tile_x=tile_col*gimp_tile_width();
	gint const m_tile_y=tile_row*gimp_tile_height();

	// if we are at the end of current relevant tile row
	if(m_tile_x+(gint)m_tile->ewidth>=m_region->x()+m_region->width()){
		// if there are no more rows of relevant tiles
		if(m_tile_y+(gint)m_tile->eheight >= m_region->y()+m_region->height()){
			return m_tile->tile_num; // no next tile to move to.
		} else {
			// move to start of next relevant tile row
			guint const new_col=m_region->x()/gimp_tile_width();
			guint const new_row=tile_row+1;

			return new_row*m_region->drawable()->ntile_cols + new_col;
		}
	} else {
		// move to next tile in the row
		guint const new_col=tile_col+1;
		guint const new_row=tile_row;
		return new_row*m_region->drawable()->ntile_cols+new_col;
	}
	
}

void pixel_region::iterator::move_to_tile(guint tilenum)
{
	guint const tile_col=tilenum%m_region->drawable()->ntile_cols;
	guint const tile_row=tilenum/m_region->drawable()->ntile_cols;
	
	m_tile=gimp_drawable_get_tile(m_region->drawable(), 
			m_region->shadow(),
			tile_row, tile_col);
	gimp_tile_ref(m_tile);
	
	// Position of new tile in the drawable
	gint const m_tile_x=tile_col*gimp_tile_width();
	gint const m_tile_y=tile_row*gimp_tile_height();

	// Get upper left corner of relevant rectangular subset of m_tile 
	gint const x1=std::max(m_region->x(), m_tile_x);
	gint const y1=std::max(m_region->y(), m_tile_y);

	// Get lower right corner of relevant rectangular subset of m_tile
	gint const x2=std::min(m_region->x()+m_region->width()-1,
			m_tile_x+(gint)m_tile->ewidth-1);
	gint const y2=std::min(m_region->y()+m_region->height()-1,
			m_tile_y+(gint)m_tile->eheight-1);
	
	assert(x1<=x2 && "Given tile does not belong to the region");
	assert(y1<=y2 && "Given tile does not belong to the region");

	if(x1==m_tile_x && x2==m_tile_x+(gint)m_tile->ewidth-1){
		// The region spans the whole tile width, which means we can
		// consider the whole tile to be one single chunk (all
		// interesting pixels in it are contiguous)
		
		// m_chunk.begin points to the pixel at (x1, y1)
		m_chunk.begin=m_tile->data+((y1%gimp_tile_height())*m_tile->ewidth+x1%gimp_tile_width())*
			m_tile->bpp;
		// m_chunk.end points beyond the pixel at (x2, y2)
		m_chunk.end=m_tile->data+((y2%gimp_tile_height())*m_tile->ewidth+x2%gimp_tile_width())*
			m_tile->bpp+1*m_tile->bpp;
		m_nr_chunks=0; // no more chunks in this tile
		m_chunk_skip=0; // should not be useful since m_nr_chunks is zero
	} else {
		// The region does span on a part of the tile width. The
		// relevant pixels in the tile are not contiguous. Will need to
		// have each line-subset of pixels as a separate chunk:
		
		// m_chunk.begin points to the pixel at (x1, y1)
		m_chunk.begin=m_tile->data+((y1%gimp_tile_height())*m_tile->ewidth+x1%gimp_tile_width())*
			m_tile->bpp;
		// m_chunk.begin points beyond the pixel at (x2, y1)
		m_chunk.end=m_tile->data+((y1%gimp_tile_height())*m_tile->ewidth+x2%gimp_tile_width())*
			m_tile->bpp+1*m_tile->bpp;
		
		// Pointer to pixel (x1, y+1) - Pointer to pixel (x1, y1)
		m_chunk_skip=m_tile->data+(((y1+1)%gimp_tile_height())*m_tile->ewidth+x1%gimp_tile_width())*
			m_tile->bpp - m_chunk.begin;

		// Number of remaining lines
		m_nr_chunks=y2-y1;
	}
}
