/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef PIXEL_REGION_HPP
#define PIXEL_REGION_HPP

#include <cstdlib>
#include <new>

#include <libgimp/gimp.h>

/// Fast and easy read and/or write access to all the pixels (in arbitrary
/// order) of a rectangular subset of a GIMP drawable.
///
/// If you need to access all the pixels of a region, regardless of the order,
/// this class implements the fastest way to do it.
///
/// Typical usage is as follows:
/// - Create an instance of pixel_region specifying the location of the
///  rectangular subset of the drawable you wish to access;
/// - Use pixel_region::begin()/end() to get a pair of pixel_region::iterators;
/// - The value returned by dereferencing pixel_region::iterators is not the
///  pixels themselves but rather instances of pixel_region::chunk which stand for
///  a whole contiguous range of pixels within the region;
/// - For each chunk you get from the iterators use
///  pixel_region::chunk::begin/end to get raw (guchar*) access to the pixels.
/// 
/// Example:
/// // Will fill region with black
/// pixel_region region(drawable, 0, 0, drawable->width, 
///    drawable->height, true, true);
/// for(pixel_region::iterator it=region.begin(); it!=region.end(); ++it){
/// 	pixel_region::chunk const ch=*it; //note: iterators point to chunks
///	std::fill(ch.begin, ch.end, 0);
/// }

class pixel_region
{
	public:
	/// A region giving access to the given subset of the given
	/// drawable
	///
	/// x, y, width, height define respectively the upper left corner x
	/// and y coordinate and the width and height of the rectangular subset
	/// of the drawable you wish to access.
	///
	/// shadow tells if you want to access the shadow tiles (i.e for writing
	/// to the drawable) or if you want direct access to the pixels (i.e for
	/// reading from the drawable)
	/// dirty is usallu equal to shadow unless you know what you're doing
	pixel_region(GimpDrawable* dr, gint x, gint y, gint w, gint h,
			bool shadow, bool dirty);

	/// Drawable being accessed
	GimpDrawable* drawable() const { return m_drawable; }
	
	/// Upper left corner X value
	gint x() const { return m_x; }
	
	/// Upper left corner Y value
	gint y() const { return m_y; }
	
	/// Width of the region
	gint width() const { return m_width; }

	/// Height of the region
	gint height() const { return m_height; }

	/// Accessing shadow tiles?
	bool shadow() const { return m_shadow; }
	
	/// Sending tiles back to core when done?
	bool dirty() const { return m_dirty; }
	/// Type of objects returned when dereferencing iterators
	struct chunk 
	{ 
		/// Start of the pixel range
		guchar* begin;

		/// End of the pixel range
		guchar *end; 
	};

	/// Used for iterating over the pixel chucks of the region
	struct iterator;
	
	/// Start of the region
	iterator begin() const;
	
	/// End of the region
	iterator end() const;
	
	private:
	/// Drawable being accessed
	GimpDrawable* m_drawable;
	
	/// Upper left corner X value
	gint m_x;

	/// Upper left corner Y value
	gint m_y;

	/// Width of the region
	gint m_width;

	/// Height of the region
	gint m_height;

	/// Access shadow tiles?
	bool m_shadow;

	/// Send tiles back to core when finished?
	bool m_dirty;
};

struct pixel_region::iterator
{
	/// Class pixel_region needs special access to be able to create
	/// iterators (outside world must rely on pixel_region for
	/// creating the iterators they'll use)
	friend class pixel_region;

	/// Access the currently pointed chunk
	chunk const& operator*() const { return m_chunk; }

	/// Access the currently pointed chunk
	chunk const* operator->() const { return &m_chunk; }

	/// Move to the next chunk
	iterator& operator++();
	
	/// Move to the next chunk
	iterator operator++(int) 
	{ 
		iterator tmp(*this);
		this->operator++();
		return tmp;
	} 

	/// Compare two iterator
	bool operator==(iterator const& lhs) const;
	/// Compare two iterator
	bool operator!=(iterator const& lhs) const { return ! this->operator==(lhs); };

	/// Copy constructor
	iterator(iterator const&);

	/// Assignment 
	iterator& operator=(iterator const& lhs)
	{
		// In place destruction, then in place copy construction
		this->iterator::~iterator();
		return *new(this) iterator::iterator(lhs);
	}

	/// Destructor
	~iterator() { gimp_tile_unref(m_tile, m_region->shadow()); }

	/// Current tile
	GimpTile* tile() const { return m_tile; }

	private:
	/// Type tag used in iterator::iterator(pixel_region, end_tag)
	enum end_tag { end };

	/// Create an iterator at the start of the given pixel region
	iterator(pixel_region const& region);

	/// Create an iterator at the end   of the given pixel region
	iterator(pixel_region const& region, end_tag);

	/// Pixel region being traversed
	pixel_region const* m_region;

	/// Current tile
	GimpTile* m_tile;

	/// Current chunk of data inside m_tile
	chunk m_chunk;

	/// Offset to add to m_chunk.begin and m_chunk.end to obtain the
	///  next chunk in m_tile (if there are still more, see
	/// m_nr_chunks)
	std::ptrdiff_t m_chunk_skip;

	/// The number of remaining chunks to read from
	/// m_tile (not including m_chunk it self)
	std::size_t m_nr_chunks;

	/// Whether this is an end iterator
	bool m_end_flag;

	/// Move to the specified tile.
	///
	/// Setup m_chunk to the first chunk in the tile, and
	/// m_chunk_skip and m_nr_chunks to their correct values.
	///
	/// Overwrites m_tile directly (call gimp_tile_unref yourself)
	void move_to_tile(guint tilenum);

	/// The tile number that follows m_tile in the region
	///
	/// Will return the same tile number as current one if it is the last
	/// one
	guint next_tile_number() const;
};

inline 
pixel_region::iterator pixel_region::begin() const
{ return pixel_region::iterator(*this); }

inline 
pixel_region::iterator pixel_region::end() const
{ return pixel_region::iterator(*this, pixel_region::iterator::end); }

#endif // PIXEL_REGION_HPP
