/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <cassert>
#include <string>
#include <numeric>
#include <sstream>
#include <iostream>
#include <vector>
#include <cmath>

#include <gtk/gtk.h>
#include "result_dialog.hpp"
#include "classification.hpp"
#include "fitness_function_repertoire.hpp"
#include "colors.hpp"
#include "load_manual_segmentation.hpp"
#include "compute_histogram.hpp"
#include "jaccard_distance.hpp"

/// Holds the evaluation of results at every stage
/// of the algorithm for each possible fitness_function
std::vector< std::vector<double> > const* fitnesses;

/// Holds all the progress bars that display the evaluation
/// of a stage of the algotithm
std::vector<GtkWidget*> * fitness_indicators;

/// The drawable of the manual segmentation (if available, otherwise it's NULL)
GimpDrawable* ground_truth_drawable;

/// The gray levels used in ground_truth_drawable (in order
std::vector<guchar>* ground_truth_levels;

/// Return value of jaccard_distance when applied on ::ground_truth_drawable (or empty)
std::vector< std::pair<int, double> >* jaccard_distance_result;

void fitness_combo_handler(GtkWidget* wid, gpointer)
{
	int const active=gtk_combo_box_get_active(GTK_COMBO_BOX(wid));
	if(active<0) return;

	for(std::vector<GtkWidget*>::iterator it=fitness_indicators->begin();
			it!=fitness_indicators->end(); ++it){
		double const fitness=(*fitnesses)[it-fitness_indicators->begin()][active];
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(*it), 
				fitness>1.0?1.0:fitness);
		std::ostringstream oss; oss.precision(4); oss<<fitness;
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR(*it), oss.str().c_str());
	}
}

/// Creates and returns a combo box containing an item per possible fitness
/// function
GtkWidget* create_fitness_combo()
{

	GtkWidget* combo=gtk_combo_box_new_text();
	fitness_function_repertoire& rep=*fitness_function_repertoire::instance();
	for(fitness_function_repertoire::iterator it=rep.begin();
			it!=rep.end(); ++it){
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo), (*it)->label());
	}
	g_signal_connect(G_OBJECT(combo), "changed", G_CALLBACK(fitness_combo_handler),
			NULL);
	return combo;
}

/// Creates and returns a widget displaying results of given stage
GtkWidget* pass_result(histogram const& hist,
		centroid_set const& centers, bool last_step=false)
{
	bool const debug=false;

	classification c12n;
	classify(hist, centers, &c12n);
	if(debug)for(classification::iterator it=c12n.begin(); it!=c12n.end(); ++it)
		std::printf("<ClsInd:%d Gray:%d> ", it->first, it->second);
	if(debug)std::printf("\n");
	/// Total number of pixels
	histogram::amount_type const nr_pixels=std::accumulate(hist.begin(),
			hist.end(), 0);
	
	enum {column_color, column_white_space,
		column_center, column_pixel_ratio, column_mean, 
		column_variance, column_root_var, column_jaccard_index,
		column_jaccard_color, nr_columns };

	// Parent of all subsequent widgets
	GtkWidget* main_box=gtk_vbox_new(false, 10);

	GtkListStore* list_store=gtk_list_store_new(nr_columns, 
			G_TYPE_STRING,
			G_TYPE_STRING,
			G_TYPE_DOUBLE,
			G_TYPE_DOUBLE, 
			G_TYPE_DOUBLE,
			G_TYPE_DOUBLE,
			G_TYPE_DOUBLE,
			G_TYPE_DOUBLE,
			G_TYPE_STRING);
	
	GtkTreeIter iter;

	for(centroid_set::const_iterator it=centers.begin();
			it!=centers.end(); ++it){
		std::pair<classification::iterator,
			classification::iterator> t=c12n.equal_range(it-centers.begin());
		
		double nr_pixel_partial=0;
		double weighted_square_sum=0;
		double weighted_sum=0;
		for(classification::iterator jt=t.first; jt!=t.second; ++jt){
			histogram::gray_level_type x=jt->second;
			histogram::amount_type n=hist[x];

			if(debug)std::printf("[Gray:%d Num:%d] ",x,n);
			
			nr_pixel_partial+=double(n);
			weighted_square_sum+=double(n)*(double(x)*double(x));
			weighted_sum+=double(n)*double(x);
		}
		double const mean=weighted_sum/nr_pixel_partial;
		double const variance=weighted_square_sum/nr_pixel_partial-mean*mean;
	
		if(debug)std::printf(" classind:%f N:%f weighted_sum:%f weight_square:%f\n",
				*it, nr_pixel_partial, 
				weighted_sum,
				weighted_square_sum);

		char char_buffer[10];
		std::sprintf(char_buffer, "#%02X%02X%02X",
				colors[it-centers.begin()].r, 
				colors[it-centers.begin()].g, 
				colors[it-centers.begin()].b);

		double jaccard_index=0.0;
		char jaccard_color[10]="#FFFFFF";
		if(::ground_truth_drawable && last_step){
			// look for jaccard index and color of class it
			for(std::vector<std::pair<int,double> >::iterator jt=
					::jaccard_distance_result->begin();
					jt!=::jaccard_distance_result->end();
					++jt){
				if(jt->first==it-centers.begin()){
					jaccard_index=jt->second;
					int color=::ground_truth_levels->operator[]
						(jt-::jaccard_distance_result->begin());
					sprintf(jaccard_color, "#%02X%02X%02X",
					      color, color, color);
					break;
				}
			}
		}

		gtk_list_store_append (list_store, &iter);
		gtk_list_store_set (list_store, &iter,
				column_center, double(*it),
				column_color, char_buffer,
				column_white_space, " ",
				column_pixel_ratio, 100*
					(double(nr_pixel_partial)/double(nr_pixels)),
				column_mean,  mean,
				column_variance, variance,
				column_root_var, std::sqrt(variance),
				column_jaccard_index, jaccard_index,
				column_jaccard_color, jaccard_color,
				
				-1);
	}

	GtkWidget* tree=gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
			gtk_tree_view_column_new_with_attributes("Colors",
				gtk_cell_renderer_text_new(),
				"text", column_white_space, 
				"background", column_color, 
				NULL));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
			gtk_tree_view_column_new_with_attributes("Centers",
				gtk_cell_renderer_text_new(),
				"text", column_center, 
				NULL));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
			gtk_tree_view_column_new_with_attributes("Pixel Coverage Ratio (%)",
				gtk_cell_renderer_text_new(),
				"text", column_pixel_ratio,  
				NULL));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
			gtk_tree_view_column_new_with_attributes("Mean",
				gtk_cell_renderer_text_new(),
				"text", column_mean, 
				NULL));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
			gtk_tree_view_column_new_with_attributes("Variance",
				gtk_cell_renderer_text_new(),
				"text", column_variance,  
				NULL));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
			gtk_tree_view_column_new_with_attributes("Root Variance",
				gtk_cell_renderer_text_new(),
				"text", column_root_var, 
				NULL));
	if(last_step && ::ground_truth_drawable){
		gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
				gtk_tree_view_column_new_with_attributes("Jaccard Index",
					gtk_cell_renderer_text_new(),
					"text", column_jaccard_index, 
					NULL));
		gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
				gtk_tree_view_column_new_with_attributes("Supervised Class",
					gtk_cell_renderer_text_new(),
					"text", column_white_space, 
					"background", column_jaccard_color, 
					NULL));
	}

	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(tree)),
			GTK_SELECTION_NONE);
	//gtk_widget_set_sensitive(tree, false);
	
	GtkWidget* progress=gtk_progress_bar_new();
	::fitness_indicators->push_back(progress);
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), 0.6);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), "60%");

	GtkWidget* hbox=gtk_hbox_new(false, 10);
	GtkWidget* label=gtk_label_new("Performance: ");
	gtk_box_pack_start(GTK_BOX(hbox), label, false, false, 0);
	gtk_box_pack_start(GTK_BOX(hbox), progress, false, false, 0);
	gtk_widget_show(progress);
	gtk_widget_show(label);

	std::ostringstream oss;
	oss<<"Number of classes: "<<centers.size();
	GtkWidget* label2=gtk_label_new(oss.str().c_str());
	gtk_misc_set_alignment(GTK_MISC(label2), 0, 0);

	gtk_box_pack_start(GTK_BOX(main_box), label2, false, false, 0); 
	gtk_box_pack_start(GTK_BOX(main_box), tree, true, true, 0); 
	gtk_box_pack_start(GTK_BOX(main_box), hbox, false, false, 0);

	gtk_widget_show(tree);
	gtk_widget_show(hbox);
	gtk_widget_show(label2);
	return main_box;
}

void result_dialog(
		gint32 image_id,
		std::vector<pass*> const& algo,
		histogram const& hist,
		std::vector< centroid_set > const& outcome)
{
	std::vector< std::vector<double> > fitnesses;
	::fitnesses=&fitnesses;
	std::vector<GtkWidget*> fitness_indicators;
	::fitness_indicators=&fitness_indicators;
	std::vector< std::pair<int,double> > jaccard_distance_result;
	::jaccard_distance_result=&jaccard_distance_result;
	std::vector<guchar> ground_truth_levels;
	::ground_truth_levels=&ground_truth_levels;

	// Load expected segmented results (if available)
	::ground_truth_drawable=load_manual_segmentation(image_id);
	if(::ground_truth_drawable){
		// Compute jaccard distance for last step
		GimpDrawable* drawable=gimp_drawable_get(gimp_image_get_active_drawable(image_id));

		classification c12n; classify(hist, outcome.back(), &c12n);

		histogram ground_hist; compute_histogram(::ground_truth_drawable, &ground_hist,
				0, 0, ::ground_truth_drawable->width,
				::ground_truth_drawable->height);

		// fill ::ground_truth_levels
		for(histogram::iterator it=ground_hist.begin(); it!=ground_hist.end(); ++it){
			if(*it) ::ground_truth_levels->push_back(it-ground_hist.begin());
		}

		jaccard_distance(drawable, c12n, outcome.back(), ::ground_truth_drawable,
				ground_hist,
				0, 0, drawable->width, drawable->height, &jaccard_distance_result);

		for(std::vector<std::pair<int, double> >::iterator it=jaccard_distance_result.begin();
				it!=jaccard_distance_result.end(); ++it)
			std::cout<<it->first<< " " <<it->second<<std::endl;


			
	}

	assert(algo.size()==outcome.size());

	GtkWidget* dialog=gtk_dialog_new_with_buttons("Results",
			NULL,//parent
			GtkDialogFlags(0),
			GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
			NULL);

	GtkWidget* vbox=gtk_vbox_new(false, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox),
				10);
	
	GtkWidget* hbox=gtk_hbox_new(false, 10);
	GtkWidget* fitness_label=gtk_label_new("Fitness Function");
	GtkWidget* fitness_box=create_fitness_combo();
	gtk_box_pack_start(GTK_BOX(hbox), fitness_label, false, false, 0);
	gtk_box_pack_start(GTK_BOX(hbox), fitness_box, false, false, 0);
	gtk_widget_show(fitness_label);
	gtk_widget_show(fitness_box);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, false, false, 10);
	gtk_widget_show(hbox);

	for(std::vector<pass*>::size_type i=0; i<algo.size(); ++i){
		GtkWidget* res=pass_result(hist, outcome[i], i==algo.size()-1);

		std::string str("After "); str+=algo[i]->label();
		GtkWidget* frame=gtk_frame_new(str.c_str());
		
		GtkWidget* exp=gtk_expander_new("Results");
		gtk_container_add(GTK_CONTAINER(exp), res);
		gtk_widget_show(res);
		
		gtk_container_add(GTK_CONTAINER(frame), exp);
		gtk_widget_show(exp);

		gtk_container_set_border_width(GTK_CONTAINER(exp), 10);
		gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 0);
		gtk_widget_show(frame);
		if(i==algo.size()-1) gtk_expander_set_expanded(GTK_EXPANDER(exp), true);

		/// Compute the value of all fitness function for this stage
		{
			classification c12n; classify(hist, outcome[i], &c12n);

			fitness_function_repertoire& rep=*fitness_function_repertoire::instance();
			fitnesses.resize(fitnesses.size()+1);
			std::vector<double>& back=fitnesses.back();
			for(fitness_function_repertoire::iterator it=
					rep.begin(); it!=rep.end(); ++it){
				back.push_back((*it)->operator() (hist,
							outcome[i], c12n));
			}
		}
	}

	GtkWidget* scrolled=gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_set_border_width(GTK_CONTAINER(scrolled), 10);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled), vbox);
	gtk_widget_show(vbox);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), scrolled, true, true, 0);
	gtk_widget_show(scrolled);
	
	gtk_combo_box_set_active(GTK_COMBO_BOX(fitness_box), 0);

	gtk_window_set_default_size(GTK_WINDOW(dialog), 700, 400);
	
	gtk_dialog_run(GTK_DIALOG(dialog));
	
	gtk_widget_destroy(dialog);
}
