/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <map>

#include "swarm_def.hpp"
#include "classification.hpp"

//parameters used to define a priority in the fitness function
double w1=0.5;
double w2=0.5;

double max_intra_distance(std::vector<double> const& centroids, histogram const& hist)
{
	typedef  std::pair<classification::iterator,
		 classification::iterator> pair;
	classification cl;
	classify(hist, centroids, &cl);
	double distance=0;
	int n=0;
	double result=0;
	for(unsigned int i=0;i<centroids.size();++i)
	{
		pair pr=cl.equal_range(i);
		for (classification::iterator it=pr.first;it!=pr.second;++it)
		{
			distance+= std::abs(it->second-centroids[i]);
			n++;
		}
		distance=distance/n;
		if(distance>result) result=distance;
		n=0;
		distance=0;
	}

	return result;
}

double min_inter_distance(std::vector<double> const& centroids)
{
	double result=255;
	double distance=0;
	for (unsigned int i=0;i<centroids.size();++i)
	{
		for(unsigned int j=i+1;j<centroids.size();++j)
		{
			distance=std::abs(centroids[i]-centroids[j]);
			if (distance <result) result=distance;
		}
	}
	return result;
}

double amrane_fitness_impl (std::vector<double> const& centroids, histogram const& hist)
{
	double result=w1*max_intra_distance(centroids, hist)+w2*(255-min_inter_distance(centroids));
	return result;
}

particle::particle (int particle_size, int velocity_range, histogram const& hist)
{

	std::vector<histogram::gray_level_type> vec; vec.reserve(256);
	for(histogram::const_iterator it=hist.begin(); it!=hist.end(); ++it)
	{
		if(*it)
			vec.push_back(it-hist.begin());
	}

	std::random_shuffle(vec.begin(), vec.end());

	std::vector<histogram::gray_level_type>::iterator it=vec.begin();
	for(int i=0; i<particle_size; ++i)
	{
		if(it==vec.end())
			break;
		current_position.push_back(*it);
		++it;
	}   
	generate_n(back_inserter(velocity),particle_size,random::random(velocity_range));
	copy(current_position.begin(), current_position.end(), back_inserter(best_position));

}

void particle::print()
{
	std::cout<<"Current position: "<<std::endl;
	for(unsigned int i=0 ;i<current_position.size(); ++i)
	{
		std::cout<<current_position[i]<<" ";
	}
	std::cout<<std::endl;
	std::cout<<"current position cost: "<<cost<<std::endl;
	std::cout<<"Best position: "<<std::endl;
	for(unsigned int i=0 ;i<best_position.size(); ++i)
	{
		std::cout<<best_position[i]<<" ";
	}
	std::cout<<std::endl;
	std::cout<<"Best position cost: "<<best_cost<<std::endl;
	std::cout<<"Velocity: "<<std::endl;
	for(unsigned int i=0 ;i<velocity.size(); ++i)
	{
		std::cout<<velocity[i]<<" ";
	}
	std::cout<<std::endl;
	std::cout<<"------------------------------------------------------"<<std::endl;

}


swarm::swarm(double _w, double _c1, double _c2, int _s, int _ps,
		int _vr, std::vector<double> & centroids, histogram const& _hist)
:w(_w), c1(_c1), c2(_c2),s(_s), ps(centroids.size()), vr(_vr),hist(_hist)

{

	//best_particle_cost=amrane_fitness_impl(best_particle_position,hist);

	for (int i=0;i<s-1;++i)
	{
		particle p(ps,vr,hist);
		p.set_cost(amrane_fitness_impl(p.get_current_position(),hist));
		p.set_best_cost(p.get_cost());
		p_repertoire.push_back(p);
	}
	particle p(ps,vr,hist);

	p.get_current_position().clear();
	p.get_best_position().clear();

	std::copy(centroids.begin(),centroids.end(),back_inserter(p.get_current_position()));
	p.set_cost(amrane_fitness_impl(p.get_current_position(),hist));
	std::copy(centroids.begin(),centroids.end(),back_inserter(p.get_best_position()));
	p.set_best_cost(p.get_cost());
	p_repertoire.push_back(p);

	generate_n(back_inserter(r1),psize(),random::random(0));
	generate_n(back_inserter(r2),psize(),random::random(0));
	best_particle_cost=110000.00; 

	//set the best particle of the swarm to the first one
	int b_part=0;
	for (unsigned int i=1;i<p_repertoire.size();++i)
	{
		if(p_repertoire[i].get_cost()<best_particle_cost)
		{
			b_part=i;
			best_particle_cost=p_repertoire[i].get_cost();
		}
	}
	best_particle_position.clear();
	copy(p_repertoire[b_part].get_best_position().begin(),
			p_repertoire[b_part].get_best_position().end(),
			back_inserter(best_particle_position));
}


void swarm::update_particle(particle& p)
{
	//Updating particle velocity

	for(int i=0;i<psize();++i)
	{
		//Velocity update function
		p.get_velocity()[i]=w*p.get_velocity()[i]+
			c1*r1[i]*(p.get_best_position()[i]-p.get_current_position()[i])+
			c2*r2[i]*(get_best_particle_position()[i]-p.get_current_position()[i]);



	}

	//updating the current position
	for(int i=0;i<psize();++i)
	{
		p.get_current_position()[i]=p.get_current_position()[i]+p.get_velocity()[i];
	}
	p.set_cost(amrane_fitness_impl(p.get_current_position(), hist));

	//Updating the best position of a particle so far
	if(p.get_cost()<p.get_best_cost())
	{
		p.get_best_position().clear();
		copy(p.get_current_position().begin(),p.get_current_position().end(),
				back_inserter(p.get_best_position()));
		p.set_best_cost(p.get_cost());
	}

}

void swarm::update_swarm()
{
	//Updating all the particles of the swarm
	for(int i=0; i<s; ++i)
	{
		update_particle(p_repertoire[i]);
	}
	int b_part=0;
	bool enter=0;
	//Updating the best particle postition( the optimal solution so far)
	for (unsigned int i=0;i<p_repertoire.size();++i)
	{
		if(p_repertoire[i].get_best_cost()<best_particle_cost)
		{
			enter=1;
			b_part=i;
			best_particle_cost=p_repertoire[i].get_best_cost();
		}
	}
	if(enter)
	{
		best_particle_position.clear();
		copy(p_repertoire[b_part].get_best_position().begin(),
				p_repertoire[b_part].get_best_position().end(),
				back_inserter(best_particle_position));
	}
}
;
