/// Multipass: A gimp plug-in for image segmentation
///
/// Copyright (C) 2007  Anis Benyelloul (anis.benyelloul@gmail.com) and
/// 			Said-Eddine Bensalem (said.bensalem@gmail.com)
///
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef SWARM_DEF_HPP
#define SWARM_DEF_HPP

#include<iostream>
#include<vector>
#include<algorithm>
#include<cmath>
#include "irandom.hpp"
#include "histogram.hpp"
#include "centroid_set.hpp"

double amrane_fitness_impl(centroid_set const& centroids, histogram const& hist);

//Represents an individual (particle) of a swarm
class particle
{
private:
    std::vector<double> current_position;//current position of a particle
    std::vector<double> best_position; //the best position reacher by a particle so far
    std::vector<double> velocity; //the speed vector of a particle
    double cost; //The fitness value of a particle
    double best_cost;
public:

    //First constructor which make a particle of particle_size size
    //velocity_range is used to initialize each component of velocity vector
    //in the rage [- velocity_range, +velocity_range]
    particle (int particle_size, int velocity_range, histogram const& hist);

    //print a particle
    void print();

    //yields the current position of the particle
    std::vector<double> & get_current_position()
    {
        return current_position;
    }

    //yields the velocity of the particle
    std::vector<double> & get_velocity()
    {
        return velocity;
    }

    //yields the best position of the particle so far
    std::vector<double> & get_best_position()
    {
        return best_position;
    }

    //sets a fitness value to a particle
    void set_cost(double c)
    {
        cost=c;
    }

    //yields a fitness value from a particle
    double const get_cost()
    {
        return cost;
    }

    //sets the best fitness value of a particle
    void set_best_cost(double c)
    {
        best_cost=c;
    }

    //gets the best fitness value from a particle
    double const get_best_cost()
    {
        return best_cost;
    }
}
;// class particle

//This class holds all the particles of the swarm and the globale characteritics
//needed for the evolutionary process
class swarm
{

private:
    std::vector<particle>  p_repertoire;
    std::vector<double> best_particle_position;
    std::vector<double> r1; //the r1->U(0,1) parameters to update particle velocity
    std::vector<double> r2; // the r2->U(0,1) parameter to update particle velocity
    double w;  //w is the inertia
    double c1,c2; // acceleration constants
    int s; //swarm size
    int ps; //particle size
    int vr; //Velocity range [-vr,+vr]
    double best_particle_cost;
    histogram const& hist;
public:

    //The arguments of the constructor are in this order: inertia, first constant,
    //second constant, swarm size, particle size, seed to initialize particles,
    //velocity range, reference to centroids vector, reference to the clustering
    //environement(histogram), the last one is a pointer to fuction to calculate
    //the fitness of a particle X[i] in the swarm; its first argument is a reference
    //to a particle which we want to calculate its fitness
    //and the last argument is a reference to the histogram.

    swarm(double inertia, double first_constant, double second_constant,
          int swarm_size, int particle_size, int velocity_range,
          std::vector<double> & centroids, histogram const& hist);

    //This member function is to update a particle of the swarm position (toward the best one)
    //according to the previous ones and to the neighbourhood
    void update_particle(particle& p);

    //This member function is to update the entire swarm
    void update_swarm();


    //Return the swarm size
    const int ssize()
    {
        return s;
    }

    //Return the particle size
    const int psize()
    {
        return ps;
    }

    //Returns a ith particles
    particle const& operator[] (int i)
    {
        return p_repertoire[i];
    }

    double const get_best_particle_cost()
    {
        return best_particle_cost;
    }

    //sets the cost of the best particle of the swarm
    void set_best_particle_cost(double b)
    {
        best_particle_cost=b;
    }

    //gets the best particle of the swarm
    std::vector<double> const& get_best_particle_position()
    {
        return best_particle_position;
    }

    //sets the best particle of the swarm
    void set_best_particle_position(std::vector<double>& p)
    {
        best_particle_position.clear();
        copy(p.begin(), p.end(), back_inserter(best_particle_position));
    }
}//class swarm
;

#endif //SWARM_DEF_HPP
